# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* Binary Execution Wrappers **************************************************************

from psyneulink.core.globals.context import Context

from collections import Counter
import copy
import ctypes
import numpy as np
from inspect import isgenerator
import itertools
import sys


from psyneulink.core import llvm as pnlvm
from . import helpers, jit_engine, builder_context
from .debug import debug_env

__all__ = ['CompExecution', 'FuncExecution', 'MechExecution']


def _convert_ctype_to_python(x):
    if isinstance(x, ctypes.Structure):
        return [_convert_ctype_to_python(getattr(x, field_name)) for field_name, _ in x._fields_]
    if isinstance(x, ctypes.Array):
        return [_convert_ctype_to_python(el) for el in x]
    if isinstance(x, (ctypes.c_double, ctypes.c_float)):
        return x.value
    if isinstance(x, (float, int)):
        return x

    assert False, "Don't know how to convert: {}".format(x)


def _tupleize(x):
    try:
        return tuple(_tupleize(y) for y in x)
    except TypeError:
        return x if x is not None else tuple()

def _element_dtype(x):
    dt = np.dtype(x)
    while dt.subdtype is not None:
        dt = dt.subdtype[0]

    return dt

def _pretty_size(size):
    units = ['B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB']
    for u in units:
        if abs(size) > 1536.0:
            size /= 1024.0
        else:
            break

    return "{:.2f} {}".format(size, u)


class CUDAExecution:
    def __init__(self, buffers=['param_struct', 'state_struct', 'out']):
        for b in buffers:
            setattr(self, "_buffer_cuda_" + b, None)
        self._uploaded_bytes = Counter()
        self._downloaded_bytes = Counter()
        self.__debug_env = debug_env

    def __del__(self):
        if "cuda_data" in self.__debug_env:
            try:
                name = self._bin_func.name
            except AttributeError:
                name = self._composition.name

            for k, v in self._uploaded_bytes.items():
                print("{} CUDA uploaded `{}': {}".format(name, k, _pretty_size(v)))
            if len(self._uploaded_bytes) > 1:
                print("{} CUDA uploaded `total': {}".format(name, _pretty_size(sum(self._uploaded_bytes.values()))))
            for k, v in self._downloaded_bytes.items():
                print("{} CUDA downloaded `{}': {}".format(name, k, _pretty_size(v)))
            if len(self._downloaded_bytes) > 1:
                print("{} CUDA downloaded `total': {}".format(name, _pretty_size(sum(self._downloaded_bytes.values()))))

    @property
    def _bin_func_multirun(self):
        # CUDA uses the same function for single and multi run
        return self._bin_func

    def _get_ctype_bytes(self, data):
        # Return dummy buffer. CUDA does not handle 0 size well.
        if ctypes.sizeof(data) == 0:
            return bytearray(b'aaaa')
        return bytearray(data)

    def upload_ctype(self, data, name='other'):
        self._uploaded_bytes[name] += ctypes.sizeof(data)
        return jit_engine.pycuda.driver.to_device(self._get_ctype_bytes(data))

    def download_ctype(self, source, ty, name='other'):
        self._downloaded_bytes[name] += ctypes.sizeof(ty)
        out_buf = bytearray(ctypes.sizeof(ty))
        jit_engine.pycuda.driver.memcpy_dtoh(out_buf, source)
        return ty.from_buffer(out_buf)

    def __get_cuda_buffer(self, struct_name):
        private_attr_name = "_buffer_cuda" + struct_name
        private_attr = getattr(self, private_attr_name)
        if private_attr is None:
            # Set private attribute to a new buffer
            private_attr = self.upload_ctype(getattr(self, struct_name), struct_name)
            setattr(self, private_attr_name, private_attr)

        return private_attr

    @property
    def _cuda_param_struct(self):
        return self.__get_cuda_buffer("_param_struct")

    @property
    def _cuda_state_struct(self):
        return self.__get_cuda_buffer("_state_struct")

    @property
    def _cuda_data_struct(self):
        return self.__get_cuda_buffer("_data_struct")

    @property
    def _cuda_conditions(self):
        return self.__get_cuda_buffer("_conditions")

    @property
    def _cuda_out(self):
        if self._buffer_cuda_out is None:
            size = ctypes.sizeof(self._vo_ty)
            self._buffer_cuda_out = jit_engine.pycuda.driver.mem_alloc(size)
        return self._buffer_cuda_out

    def cuda_execute(self, variable):
        # Create input argument
        new_var = np.asfarray(variable, dtype=self._vi_dty)
        data_in = jit_engine.pycuda.driver.In(new_var)
        self._uploaded_bytes['input'] += new_var.nbytes

        self._bin_func.cuda_call(self._cuda_param_struct,
                                 self._cuda_state_struct,
                                 data_in, self._cuda_out,
                                 threads=len(self._execution_contexts))

        # Copy the result from the device
        ct_res = self.download_ctype(self._cuda_out, self._vo_ty, 'result')
        return _convert_ctype_to_python(ct_res)


class FuncExecution(CUDAExecution):

    def __init__(self, component, execution_ids=[None], *, tags=frozenset()):
        super().__init__()
        self._bin_func = pnlvm.LLVMBinaryFunction.from_obj(component, tags=tags)
        self._execution_contexts = [
            Context(execution_id=eid) for eid in execution_ids
        ]
        self._component = component
        self.__debug_env = debug_env

        par_struct_ty, ctx_struct_ty, vi_ty, vo_ty = self._bin_func.byref_arg_types

        if len(execution_ids) > 1:
            self._bin_multirun = self._bin_func.get_multi_run()
            self._ct_len = ctypes.c_int(len(execution_ids))
            vo_ty = vo_ty * len(execution_ids)
            vi_ty = vi_ty * len(execution_ids)

            self.__param_struct = None
            self.__state_struct = None

        self._vo_ty = vo_ty
        self._ct_vo = vo_ty()
        self._vi_ty = vi_ty
        self._vi_dty = _element_dtype(vi_ty)
        if "stat" in self.__debug_env:
            print("Input struct size:", _pretty_size(ctypes.sizeof(vi_ty)),
                  "for", self._component.name)
            print("Output struct size:", _pretty_size(ctypes.sizeof(vo_ty)),
                  "for", self._component.name)

    def _get_compilation_param(self, name, initializer, arg, context):
        param = getattr(self._component._compilation_data, name)
        struct = param._get(context)
        if struct is None:
            initializer = getattr(self._component, initializer)(context)
            struct_ty = self._bin_func.byref_arg_types[arg]
            struct = struct_ty(*initializer)
            param._set(struct, context=context)
            if "stat" in self.__debug_env:
                print("Instantiated struct:", name, "( size:" ,
                      _pretty_size(ctypes.sizeof(struct_ty)), ")",
                      "for", self._component.name)

        return struct

    def _get_multirun_struct(self, arg, init):
        struct_ty = self._bin_multirun.byref_arg_types[arg] * len(self._execution_contexts)
        initializer = (getattr(self._component, init)(ex_id) for ex_id in self._execution_contexts)
        return struct_ty(*initializer)

    @property
    def _param_struct(self):
        if len(self._execution_contexts) > 1:
            if self.__param_struct is None:
                self.__param_struct = self._get_multirun_struct(0, '_get_param_initializer')
            return self.__param_struct

        return self._get_compilation_param('parameter_struct', '_get_param_initializer', 0, self._execution_contexts[0])

    @property
    def _state_struct(self):
        if len(self._execution_contexts) > 1:
            if self.__state_struct is None:
                self.__state_struct = self._get_multirun_struct(1, '_get_state_initializer')
            return self.__state_struct

        return self._get_compilation_param('state_struct', '_get_state_initializer', 1, self._execution_contexts[0])

    def execute(self, variable):
        # Make sure function inputs are 2d.
        # Mechanism inptus are already 3d so the first part is nop.
        new_variable = np.asfarray(np.atleast_2d(variable),
                                   dtype=self._vi_dty)

        if len(self._execution_contexts) > 1:
            # wrap_call casts the arguments so we only need contiguous data
            # layout
            ct_vi = np.ctypeslib.as_ctypes(new_variable)
            self._bin_multirun.wrap_call(self._param_struct,
                                         self._state_struct,
                                         ct_vi, self._ct_vo, self._ct_len)
        else:
            ct_vi = np.ctypeslib.as_ctypes(new_variable)
            self._bin_func(ctypes.byref(self._param_struct),
                           ctypes.byref(self._state_struct),
                           ct_vi, ctypes.byref(self._ct_vo))

        return _convert_ctype_to_python(self._ct_vo)


class MechExecution(FuncExecution):

    def execute(self, variable):
        # Convert to 3d. We always assume that:
        #   a) the input is vector of input ports
        #   b) input ports take vector of projection outputs
        #   c) projection output is a vector (even 1 element vector)
        new_var = np.atleast_3d(variable)
        new_var.shape = (len(self._component.input_ports), 1, -1)
        return super().execute(new_var)


class CompExecution(CUDAExecution):

    def __init__(self, composition, execution_ids=[None], *, additional_tags=frozenset()):
        super().__init__(buffers=['state_struct', 'param_struct', 'data_struct', 'conditions'])
        self._composition = composition
        self._execution_contexts = [
            Context(execution_id=eid) for eid in execution_ids
        ]
        self.__bin_exec_func = None
        self.__bin_exec_multi_func = None
        self.__bin_func = None
        self.__bin_run_func = None
        self.__bin_run_multi_func = None
        self.__debug_env = debug_env
        self.__frozen_vals = None
        self.__tags = frozenset(additional_tags)

        # TODO: Consolidate these
        if len(execution_ids) > 1:
            self.__state_struct = None
            self.__param_struct = None
            self.__data_struct = None
            self.__conds = None
            self._ct_len = ctypes.c_int(len(execution_ids))

    @property
    def _bin_func(self):
        if self.__bin_func is not None:
            assert len(self._execution_contexts) == 1
            return self.__bin_func
        if self.__bin_exec_func is not None:
            return self.__bin_exec_func
        if self.__bin_run_func is not None:
            return self.__bin_run_func

        assert False, "Binary function not set for execution!"

    @property
    def _bin_func_multirun(self):
        if self.__bin_exec_multi_func is not None:
            return self.__bin_exec_multi_func
        if self.__bin_run_multi_func is not None:
            return self.__bin_run_multi_func

        return super()._bin_func_multirun

    def _set_bin_node(self, node):
        assert node in self._composition._all_nodes
        wrapper = builder_context.LLVMBuilderContext.get_global().get_node_wrapper(self._composition, node)
        self.__bin_func = pnlvm.LLVMBinaryFunction.from_obj(
            wrapper, tags=self.__tags.union({"node_wrapper"}))

    @property
    def _conditions(self):
        if len(self._execution_contexts) > 1:
            if self.__conds is None:
                cond_type = self._bin_func_multirun.byref_arg_types[4] * len(self._execution_contexts)
                gen = helpers.ConditionGenerator(None, self._composition)
                cond_initializer = (gen.get_condition_initializer() for _ in self._execution_contexts)
                self.__conds = cond_type(*cond_initializer)
            return self.__conds

        conds = self._composition._compilation_data.scheduler_conditions._get(self._execution_contexts[0])
        if conds is None:
            cond_type = self._bin_func.byref_arg_types[4]
            gen = helpers.ConditionGenerator(None, self._composition)
            cond_initializer = gen.get_condition_initializer()
            conds = cond_type(*cond_initializer)
            self._composition._compilation_data.scheduler_conditions._set(conds, context=self._execution_contexts[0])
            if "stat" in self.__debug_env:
                print("Instantiated condition struct ( size:" ,
                      _pretty_size(ctypes.sizeof(struct_ty)), ")",
                      "for", self._composition.name)
        return conds

    def _get_compilation_param(self, name, initializer, arg, context):
        param = getattr(self._composition._compilation_data, name)
        struct = param._get(context)
        if struct is None:
            initializer = getattr(self._composition, initializer)(context)
            struct_ty = self._bin_func.byref_arg_types[arg]
            struct = struct_ty(*initializer)
            param._set(struct, context=context)
            if "stat" in self.__debug_env:
                print("Instantiated struct:", name, "( size:" ,
                      _pretty_size(ctypes.sizeof(struct_ty)), ")",
                      "for", self._composition.name)

        return struct

    def _get_multirun_struct(self, arg, init):
        struct_ty = self._bin_func_multirun.byref_arg_types[arg] * len(self._execution_contexts)
        initializer = (getattr(self._composition, init)(ex) for ex in self._execution_contexts)
        return struct_ty(*initializer)

    @property
    def _param_struct(self):
        if len(self._execution_contexts) > 1:
            if self.__param_struct is None:
                self.__param_struct = self._get_multirun_struct(1, '_get_param_initializer')
            return self.__param_struct

        return self._get_compilation_param('parameter_struct', '_get_param_initializer', 1, self._execution_contexts[0])

    def _copy_params_to_pnl(self, context=None, component=None, params=None):
        # need to special case compositions
        from psyneulink.core.compositions import Composition
        from psyneulink.core.components.projections.pathway import MappingProjection

        if component is None:
            component = self._composition

        if params is None:
            assert component == self._composition
            params = self._param_struct

        if isinstance(component, Composition):
            # first handle all inner projections
            params_projections_list = getattr(params, params._fields_[1][0])
            for idx, projection in enumerate(component._inner_projections):
                projection_params = getattr(params_projections_list, params_projections_list._fields_[idx][0])
                self._copy_params_to_pnl(context=context, component=projection, params=projection_params)

            # now recurse on all nodes
            params_node_list = getattr(params, params._fields_[0][0])
            for idx, node in enumerate(component._all_nodes):
                node_params = getattr(params_node_list, params_node_list._fields_[idx][0])
                self._copy_params_to_pnl(context=context, component=node, params=node_params)
        elif isinstance(component, MappingProjection):
            # we copy all ids back
            for idx, attribute in enumerate(component.llvm_param_ids):
                to_set = getattr(component.parameters, attribute)
                parameter_ctype = getattr(params, params._fields_[idx][0])
                value = _convert_ctype_to_python(parameter_ctype)
                if attribute == 'matrix':
                    # special case since we have to unflatten matrix
                    # FIXME: this seems to break something when generalized for all attributes
                    value = np.array(value).reshape(component.parameters.matrix._get(context).shape)
                    to_set._set(value, context=context)

    @property
    def _state_struct(self):
        if len(self._execution_contexts) > 1:
            if self.__state_struct is None:
                self.__state_struct = self._get_multirun_struct(0, '_get_state_initializer')
            return self.__state_struct

        return self._get_compilation_param('state_struct', '_get_state_initializer', 0, self._execution_contexts[0])

    @property
    def _data_struct(self):
        # Run wrapper changed argument order
        arg = 2 if self._bin_func is self.__bin_run_func else 3

        if len(self._execution_contexts) > 1:
            if self.__data_struct is None:
                self.__data_struct = self._get_multirun_struct(arg, '_get_data_initializer')
            return self.__data_struct

        return self._get_compilation_param('data_struct', '_get_data_initializer', arg, self._execution_contexts[0])

    @_data_struct.setter
    def _data_struct(self, data_struct):
        if len(self._execution_contexts) > 1:
            self.__data_struct = data_struct
        else:
            self._composition._compilation_data.data_struct._set(data_struct, context=self._execution_contexts[0])

    def _extract_node_struct(self, node, data):
        # context structure consists of a list of node contexts,
        #   followed by a list of projection contexts; get the first one
        # parameter structure consists of a list of node parameters,
        #   followed by a list of projection parameters; get the first one
        # output structure consists of a list of node outputs,
        #   followed by a list of nested data structures; get the first one
        field = data._fields_[0][0]
        res_struct = getattr(data, field)
        index = self._composition._get_node_index(node)
        field = res_struct._fields_[index][0]
        res_struct = getattr(res_struct, field)
        return _convert_ctype_to_python(res_struct)

    def extract_node_struct(self, node, struct):
        if len(self._execution_contexts) > 1:
            return [self._extract_node_struct(node, struct[i]) for i, _ in enumerate(self._execution_contexts)]
        else:
            return self._extract_node_struct(node, struct)

    def extract_frozen_node_output(self, node):
        return self.extract_node_struct(node, self.__frozen_vals)

    def extract_node_output(self, node):
        return self.extract_node_struct(node, self._data_struct)

    def extract_node_state(self, node):
        return self.extract_node_struct(node, self._state_struct)

    def extract_node_params(self, node):
        return self.extract_node_struct(node, self._param_struct)

    def insert_node_output(self, node, data):
        my_field_name = self._data_struct._fields_[0][0]
        my_res_struct = getattr(self._data_struct, my_field_name)
        index = self._composition._get_node_index(node)
        node_field_name = my_res_struct._fields_[index][0]
        setattr(my_res_struct, node_field_name, _tupleize(data))

    def _get_input_struct(self, inputs):
        # Either node or composition execute.
        # All execute functions expect inputs to be 3rd param.
        c_input = self._bin_func.byref_arg_types[2]

        # Read provided input data and parse into an array (generator)
        if len(self._execution_contexts) > 1:
            assert len(self._execution_contexts) == len(inputs)
            c_input = c_input * len(self._execution_contexts)
            input_data = (([x] for x in self._composition._build_variable_for_input_CIM(inp)) for inp in inputs)
        else:
            input_data = ([x] for x in self._composition._build_variable_for_input_CIM(inputs))

        if "stat" in self.__debug_env:
            print("Input struct size:", _pretty_size(ctypes.sizeof(c_input)),
                  "for", self._composition.name)
        return c_input(*_tupleize(input_data))

    def freeze_values(self):
        self.__frozen_vals = copy.deepcopy(self._data_struct)

    def execute_node(self, node, inputs=None):
        # We need to reconstruct the input dictionary here if it was not provided.
        # This happens during node execution of nested compositions.
        assert len(self._execution_contexts) == 1
        if inputs is None and node is self._composition.input_CIM:
            context = self._execution_contexts[0]
            port_inputs = {origin_port:[proj.parameters.value._get(context) for proj in p[0].path_afferents] for (origin_port, p) in self._composition.input_CIM_ports.items()}
            inputs = {}
            for p, v in port_inputs.items():
                data = inputs.setdefault(p.owner, [0] * len(p.owner.input_ports))
                index = p.owner.input_ports.index(p)
                data[index] = v[0]


        # Set bin node to make sure self._*struct works as expected
        self._set_bin_node(node)
        if inputs is not None:
            inputs = self._get_input_struct(inputs)

        assert inputs is not None or node is not self._composition.input_CIM

        # Freeze output values if this is the first time we need them
        if node is not self._composition.input_CIM and self.__frozen_vals is None:
            self.freeze_values()

        self._bin_func(self._state_struct, self._param_struct,
                       inputs, self.__frozen_vals, self._data_struct)

        if "comp_node_debug" in self.__debug_env:
            print("RAN: {}. CTX: {}".format(node, self.extract_node_state(node)))
            print("RAN: {}. Params: {}".format(node, self.extract_node_params(node)))
            print("RAN: {}. Results: {}".format(node, self.extract_node_output(node)))

    @property
    def _bin_exec_func(self):
        if self.__bin_exec_func is None:
            self.__bin_exec_func = pnlvm.LLVMBinaryFunction.from_obj(
                self._composition, tags=self.__tags)

        return self.__bin_exec_func

    @property
    def _bin_exec_multi_func(self):
        if self.__bin_exec_multi_func is None:
            self.__bin_exec_multi_func = self._bin_exec_func.get_multi_run()

        return self.__bin_exec_multi_func

    def execute(self, inputs):
        # NOTE: Make sure that input struct generation is inlined.
        # We need the binary function to be setup for it to work correctly.
        if len(self._execution_contexts) > 1:
            self._bin_exec_multi_func.wrap_call(self._state_struct,
                                                self._param_struct,
                                                self._get_input_struct(inputs),
                                                self._data_struct,
                                                self._conditions, self._ct_len)
        else:
            self._bin_exec_func(self._state_struct, self._param_struct,
                                self._get_input_struct(inputs),
                                self._data_struct, self._conditions)

    def cuda_execute(self, inputs):
        # NOTE: Make sure that input struct generation is inlined.
        # We need the binary function to be setup for it to work correctly.
        self._bin_exec_func.cuda_call(self._cuda_state_struct,
                                      self._cuda_param_struct,
                                      self.upload_ctype(self._get_input_struct(inputs), 'input'),
                                      self._cuda_data_struct,
                                      self._cuda_conditions,
                                      threads=len(self._execution_contexts))

        # Copy the data struct from the device
        self._data_struct = self.download_ctype(self._cuda_data_struct, type(self._data_struct), '_data_struct')

    # Methods used to accelerate "Run"

    def _get_run_input_struct(self, inputs, num_input_sets):
        input_type = self._bin_run_func.byref_arg_types[3]
        c_input = (input_type * num_input_sets) * len(self._execution_contexts)
        if len(self._execution_contexts) == 1:
            inputs = [inputs]

        assert len(inputs) == len(self._execution_contexts)
        # Extract input for each trial and execution id
        run_inputs = ((([x] for x in self._composition._build_variable_for_input_CIM({k:v[i] for k,v in inp.items()})) for i in range(num_input_sets)) for inp in inputs)
        return c_input(*_tupleize(run_inputs))

    def _get_generator_run_input_struct(self, inputs, runs):
        assert len(self._execution_contexts) == 1
        # Extract input for each trial
        run_inputs = ((np.atleast_2d(x) for x in self._composition._build_variable_for_input_CIM({k:np.atleast_1d(v) for k,v in inp.items()})) for inp in inputs)
        run_inputs = _tupleize(run_inputs)
        num_input_sets = len(run_inputs)
        runs = num_input_sets if runs == 0 or runs == sys.maxsize else runs
        c_input = self._bin_run_func.byref_arg_types[3] * num_input_sets
        return c_input(*run_inputs), runs

    @property
    def _bin_run_func(self):
        if self.__bin_run_func is None:
            self.__bin_run_func = pnlvm.LLVMBinaryFunction.from_obj(
                self._composition, tags=self.__tags.union({"run"}))

        return self.__bin_run_func

    @property
    def _bin_run_multi_func(self):
        if self.__bin_run_multi_func is None:
            self.__bin_run_multi_func = self._bin_run_func.get_multi_run()

        return self.__bin_run_multi_func

    def run(self, inputs, runs=0, num_input_sets=0):
        if isgenerator(inputs):
            inputs, runs = self._get_generator_run_input_struct(inputs, runs)
            assert num_input_sets == 0 or num_input_sets == sys.maxsize
            num_input_sets = len(inputs)
        else:
            inputs = self._get_run_input_struct(inputs, num_input_sets)

        ct_vo = self._bin_run_func.byref_arg_types[4] * runs
        if len(self._execution_contexts) > 1:
            ct_vo = ct_vo * len(self._execution_contexts)
        outputs = ct_vo()

        if "stat" in self.__debug_env:
            print("Input struct size:", _pretty_size(ctypes.sizeof(inputs)),
                  "for", self._composition.name)
            print("Output struct size:", _pretty_size(ctypes.sizeof(outputs)),
                  "for", self._composition.name)

        runs_count = ctypes.c_int(runs)
        input_count = ctypes.c_int(num_input_sets)
        if len(self._execution_contexts) > 1:
            self._bin_run_multi_func.wrap_call(self._state_struct, self._param_struct,
                                               self._data_struct, inputs, outputs,
                                               runs_count, input_count, self._ct_len)
        else:
            self._bin_run_func.wrap_call(self._state_struct, self._param_struct,
                                         self._data_struct, inputs, outputs,
                                         runs_count, input_count)
        return _convert_ctype_to_python(outputs)

    def cuda_run(self, inputs, runs, num_input_sets):
        # Create input buffer
        if isgenerator(inputs):
            inputs, runs = self._get_generator_run_input_struct(inputs, runs)
            assert num_input_sets == 0 or num_input_sets == sys.maxsize
            num_input_sets = len(inputs)
        else:
            inputs = self._get_run_input_struct(inputs, num_input_sets)
        data_in = self.upload_ctype(inputs, 'input')

        # Create output buffer
        output_type = (self._bin_run_func.byref_arg_types[4] * runs)
        if len(self._execution_contexts) > 1:
            output_type = output_type * len(self._execution_contexts)
        output_size = ctypes.sizeof(output_type)
        data_out = jit_engine.pycuda.driver.mem_alloc(output_size)

        runs_count = jit_engine.pycuda.driver.In(np.int32(runs))
        input_count = jit_engine.pycuda.driver.In(np.int32(num_input_sets))
        # runs_count + input_count
        self._uploaded_bytes['input'] += 8

        self._bin_run_func.cuda_call(self._cuda_state_struct,
                                     self._cuda_param_struct,
                                     self._cuda_data_struct,
                                     data_in, data_out, runs_count, input_count,
                                     threads=len(self._execution_contexts))

        # Copy the data struct from the device
        ct_out = self.download_ctype(data_out, output_type, 'result')
        return _convert_ctype_to_python(ct_out)

    def cuda_evaluate(self, variable, search_space):
        ocm = self._composition.controller
        assert len(self._execution_contexts) == 1
        context = self._execution_contexts[0]

        bin_func = pnlvm.LLVMBinaryFunction.from_obj(ocm, tags=frozenset({"evaluate"}))
        self.__bin_func = bin_func
        assert len(bin_func.byref_arg_types) == 6

        # There are 6 arguments to evaluate:
        # comp_param, comp_state, allocations, results, output, input, comp_data
        # all but #2 and #3 are shared

        # Directly initialized structures
        ct_comp_param = bin_func.byref_arg_types[0](*ocm.agent_rep._get_param_initializer(context))
        ct_comp_state = bin_func.byref_arg_types[1](*ocm.agent_rep._get_state_initializer(context))
        ct_comp_data = bin_func.byref_arg_types[5](*ocm.agent_rep._get_data_initializer(context))

        # Construct the allocations array
        alloc_vals = itertools.product(*search_space)
        alloc_dty = _element_dtype(bin_func.byref_arg_types[2])
        allocations = np.asfarray(np.atleast_2d([*alloc_vals]), dtype=alloc_dty)
        ct_allocations = allocations.ctypes.data_as(ctypes.POINTER(bin_func.byref_arg_types[2] * len(allocations)))

        # Construct input variable
        el_dty = _element_dtype(bin_func.byref_arg_types[4])
        converted_variable = np.array(np.asfarray(x, dtype=el_dty) for x in variable)
        ct_in = converted_variable.ctypes.data_as(ctypes.POINTER(bin_func.byref_arg_types[4]))

        # Ouput is allocated on device, but we need the ctype.
        out_ty = bin_func.byref_arg_types[3] * len(allocations)

        cuda_args = (self.upload_ctype(ct_comp_param, 'params'),
                     self.upload_ctype(ct_comp_state, 'state'),
                     self.upload_ctype(ct_allocations.contents, 'input'),
                     jit_engine.pycuda.driver.mem_alloc(ctypes.sizeof(out_ty)),
                     self.upload_ctype(ct_in.contents, 'input'),
                     self.upload_ctype(ct_comp_data, 'data'),
                    )

        bin_func.cuda_call(*cuda_args, threads=len(allocations))
        ct_results = self.download_ctype(cuda_args[3], out_ty, 'result')

        return ct_allocations.contents, ct_results



from psyneulink import *

ia = ProcessingMechanism(name='INNER INPUT')
ib = ProcessingMechanism(name='INNER OUTPUT')
input_mech = ProcessingMechanism(name='OUTER INPUT')
internal_mech = ProcessingMechanism(name='INTERNAL')
output_mech = ProcessingMechanism(name='OUTER OUTPUT')
ctl_mech = ControlMechanism(name='CONTROL',
                            control=[(SLOPE, input_mech),
                                     (SLOPE, output_mech)])
target = ProcessingMechanism(name='TARGET')
icomp = Composition(name="NESTED COMPOSITION")
p = icomp.add_backpropagation_learning_pathway(pathway=[ia, ib])
ocomp = Composition(name='COMPOSITION',
                    pathways=[
                        [input_mech,
                        internal_mech,
                        icomp,
                        output_mech],
                        # [target,p.target]
                    ],
                    controller=OptimizationControlMechanism(
                        name='CONTROLLER',
                        objective_mechanism=ObjectiveMechanism(name='OBJECTIVE MECHANISM',
                                                               monitor=[input_mech,
                                                                        output_mech]),
                        control=(SLOPE, internal_mech))
                    )

ocomp.show_graph(show_nested=NESTED,
                 show_learning=True,
                 # show_cim=True
                 )


# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#
#
# ******************************************  ShellClasses *************************************************************

"""Define globally-visible classes for Base classes and typechecking functions for parameters of PsyNeuLink Components

Shell Classes
-------------

Used to allow classes to refer to one another without creating import loops,
including (but not restricted to) the following dependencies:
- `Projection <Projection>` subclasses must see (particular) `Port <Port>` subclasses in order to assign
  `kwProjectionSender`
- `Port <Port>` subclasses must see (particular) `Projection <Projection>` subclasses in order to assign
  `PROJECTION_TYPE`
- `Process` must see `Mechanism <Mechanism>` subclasses to assign `PsyNeuLink.Components.DefaultMechanism`

TBI:
  `Mechanism <Mechanism>`, `Projection <Projection>` (and possibly `Port <Port>`) classes should be extensible:
  developers should be able to create, register and refer to subclasses (plug-ins), without modifying core code

"""

from psyneulink.core.components.component import Component

__all__ = [
    'Function', 'Mechanism', 'Process_Base', 'Projection', 'ShellClass', 'ShellClassError', 'Port', 'System_Base',
]


class ShellClassError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


def _attempt_to_call_base_class(cls, alternative):
    raise ShellClassError("Can't call {} directly;  must use {}".format(cls.__class__.__name__, alternative))


class ShellClass(Component):
    pass


# ******************************************* SYSTEM *******************************************************************

class Composition_Base(ShellClass):
    pass


# ******************************************* SYSTEM *******************************************************************

class System_Base(ShellClass):
    pass


# ****************************************** PROCESS *******************************************************************


class Process_Base(ShellClass):
    pass

# ******************************************* MECHANISM ****************************************************************


class Mechanism(ShellClass):

    def __init__(self,
                 default_variable=None,
                 size=None,
                 function=None,
                 param_defaults=None,
                 name=None,
                 prefs=None,
                 **kwargs):
        super().__init__(default_variable=default_variable,
                         size=size,
                         function=function,
                         param_defaults=param_defaults,
                         name=name,
                         prefs=prefs,
                         **kwargs)

    def _validate_params(self, request_set, target_set=None, context=None):
        raise ShellClassError("Must implement _validate_params in {0}".format(self))

    def adjust_function(self, params, context):
        raise ShellClassError("Must implement adjust_function in {0}".format(self))


# ********************************************* PORT ******************************************************************


class Port(ShellClass):

    @property
    def owner(self):
        raise ShellClassError("Must implement @property owner method in {0}".format(self.__class__.__name__))

    @owner.setter
    def owner(self, assignment):
        raise ShellClassError("Must implement @owner.setter method in {0}".format(self.__class__.__name__))

    def _validate_variable(self, variable, context=None):
        raise ShellClassError("Must implement _validate_variable in {0}".format(self))

    def _validate_params(self, request_set, target_set=None, context=None):
        raise ShellClassError("Must implement _validate_params in {0}".format(self))

    def add_observer_for_keypath(self, object, keypath):
        raise ShellClassError("Must implement add_observer_for_keypath in {0}".format(self.__class__.__name__))

    def set_value(self, new_value):
        raise ShellClassError("Must implement set_value in {0}".format(self.__class__.__name__))

    def _update(self, params=None, context=None):
        raise ShellClassError("{} must implement update".format(self.__class__.__name__))


# ******************************************* PROJECTION ***************************************************************


class Projection(ShellClass):

    # def assign_states(self):
    #     raise ShellClassError("Must implement assign_states in {0}".format(self.__class__.__name__))
    def validate_states(self):
        raise ShellClassError("Must implement validate_states in {0}".format(self.__class__.__name__))

    def _validate_params(self, request_set, target_set=None, context=None):
        raise ShellClassError("Must implement _validate_params in {0}".format(self.__class__.__name__))


# *********************************************  FUNCTION  *************************************************************


class Function(ShellClass):

    def _instantiate_function(self, function, function_params=None, context=None):
        return



#!/bin/env python3

# Short example/demonstrator how to create and compile native code using
# LLVM MCJIT just-in-time compiler

from llvmlite import binding,ir
import ctypes
import pytest
try:
    import pycuda
    from pycuda import autoinit as pycuda_default
    # Import this after pycuda since only cuda test needs numpy
    import numpy as np
except:
    pycuda = None

@pytest.mark.llvm
def test_llvm_lite():
    # Create some useful types
    double = ir.DoubleType()
    fnty = ir.FunctionType(double, (double, double))

    # Create an empty module...
    module = ir.Module(name=__file__)
    # and declare a function named "fpadd" inside it
    func = ir.Function(module, fnty, name="fpadd")

    # Now implement basic addition
    # basic blocks are sequences of instructions that have exactly one
    # entry point and one exit point (no control flow)
    # We only need one in this case
    # See available operations at:
    # http://llvmlite.readthedocs.io/en/latest/ir/builder.html#instruction-building
    block = func.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)
    a, b = func.args
    result = builder.fadd(a, b, name="res")
    builder.ret(result)

    # Uncomment to print the module IR. This prints LLVM IR assembly.
    # print("LLVM IR:")
    # print(module)

    binding.initialize()

    # native == currently running CPU
    binding.initialize_native_target()

    # TODO: This prevents 'LLVM ERROR: Target does not support MC emission!',
    # but why?
    binding.initialize_native_asmprinter()

    # Create compilation target, use default triple
    target = binding.Target.from_default_triple()
    target_machine = target.create_target_machine()

    # And an execution engine with an empty backing module
    # TODO: why is empty backing mod necessary?
    backing_mod = binding.parse_assembly("")

    # There are other engines beside MCJIT
    # MCJIT makes it easier to run the compiled function right away.
    engine = binding.create_mcjit_compiler(backing_mod, target_machine)

    # IR module is not the same as binding module.
    # "assembly" in this case is LLVM IR assembly
    # TODO is there a better way to convert this?
    mod = binding.parse_assembly(str(module))
    mod.verify()

    # Now add the module and make sure it is ready for execution
    engine.add_module(mod)
    engine.finalize_object()

    # Uncomment to print generated x86 assembly
    #print("x86 assembly:")
    #print(target_machine.emit_assembly(mod))

    # Look up the function pointer (a Python int)
    # func_ptr is now an address to a compiled function
    func_ptr = engine.get_function_address("fpadd")

    # Run the function via ctypes
    a = 10.0
    b = 3.5
    cfunc = ctypes.CFUNCTYPE(ctypes.c_double, ctypes.c_double, ctypes.c_double)(func_ptr)
    res = cfunc(10.0, 3.5)
    assert res == (a + b)
    if res != (a + b):
        print("TEST FAILED! {} instead of {}".format(res, a + b))
    else:
        print("TEST PASSED! {} == {}".format(res, a + b))

    engine.remove_module(mod)
    # TODO: shutdown cleanly
    # we need to do something extra before shutdown
    #binding.shutdown()


@pytest.mark.llvm
@pytest.mark.cuda
@pytest.mark.skipif(pycuda is None, reason="pyCUDA modeule is not available")
def test_llvm_lite_ptx_pycuda():
    # Create some useful types
    double = ir.DoubleType()
    fnty = ir.FunctionType(ir.VoidType(), (double, double, double.as_pointer()))

    # Create an empty module...
    module = ir.Module(name=__file__)
    # and declare a function named "fpadd" inside it
    func = ir.Function(module, fnty, name="fpadd")

    # Now implement basic addition
    # basic blocks are sequences of instructions that have exactly one
    # entry point and one exit point (no control flow)
    # We only need one in this case
    # See available operations at:
    # http://llvmlite.readthedocs.io/en/latest/ir/builder.html#instruction-building
    block = func.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)
    a, b, res = func.args
    result = builder.fadd(a, b, name="res")
    builder.store(result, res)
    builder.ret_void()

    # Add kernel mark metadata
    module.add_named_metadata("nvvm.annotations",[func, "kernel", ir.IntType(32)(1)])

    # Uncomment to print the module IR. This prints LLVM IR assembly.
#    print("LLVM IR:\n", module)

    binding.initialize()

    binding.initialize_all_targets()
    binding.initialize_all_asmprinters()

    capability = pycuda_default.device.compute_capability()

    # Create compilation target, use default triple
    target = binding.Target.from_triple("nvptx64-nvidia-cuda")
    target_machine = target.create_target_machine(cpu="sm_{}{}".format(capability[0], capability[1]), codemodel='small')

    mod = binding.parse_assembly(str(module))
    mod.verify()
    ptx = target_machine.emit_assembly(mod)

    # Uncomment to print generated x86 assembly
#    print("PTX assembly:\n", ptx)

    ptx_mod = pycuda.driver.module_from_buffer(ptx.encode())
    cuda_func = ptx_mod.get_function('fpadd')

    # Run the function via ctypes
    a = np.float64(10.0)
    b = np.float64(3.5)
    res = np.empty(1, dtype=np.float64)
    dev_res = pycuda.driver.Out(res)
    cuda_func(a, b, dev_res, block=(1,1,1))
    assert res[0] == (a + b)



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# *******************************************  InputPort *****************************************************
#
"""
Contents
--------

* `InputPort_Overview`
* `InputPort_Creation`
    - `InputPort_Deferred_Initialization`
    - `InputPort_Primary`
    - `InputPort_Specification`
        • `Forms of Specification <InputPort_Forms_of_Specification>`
        • `Variable, Value and Mechanism <InputPort_Variable_and_Value>`
        • `Variable: Compatability and Constaints <InputPort_Compatability_and_Constraints>`
* `InputPort_Structure`
    - `Afferent Projections <InputPort_Afferent_Projections>`
    - `Variable <InputPort_Variable>`
    - `Function <InputPort_Function>`
    - `Value <InputPort_Value>`
    - `Weights ane Exponents <InputPort_Weights_And_Exponents>`
* `InputPort_Execution`
* `InputPort_Class_Reference`

.. _InputPort_Overview:

Overview
--------

The purpose of an InputPort is to receive and combine inputs to a `Mechanism <Mechanism>`, allow them to be modified,
and provide them to the Mechanism's `function <Mechanism_Base.function>`. An InputPort receives input to a Mechanism
provided by the `Projections <Projection>` to that Mechanism from others in a `Composition`.  If the InputPort belongs
to an `ORIGIN` Mechanism (see `Mechanism_Role_In_Compositions`), then it receives the input specified when that
Composition is `run <Run>`.  The `PathwayProjections <PathWayProjection>` received by an InputPort are listed in its
`path_afferents <Port.path_afferents>`, and its `ModulatoryProjections <ModulatoryProjection>` in its `mod_afferents
<Port.mod_afferents>` attribute.  Its `function <InputPort.function>` combines the values received from its
PathWayProjections, modifies the combined value according to value(s) any ModulatoryProjections it receives, and
provides the result to the assigned item of its owner Mechanism's `variable <Mechanism_Base.variable>` and
`input_values <Mechanism_Base.input_values>` attributes (see `below` and `Mechanism InputPorts <Mechanism_InputPorts>`
for additional details about the role of InputPorts in Mechanisms, and their assignment to the items of a Mechanism's
`variable <Mechanism_Base.variable>` attribute).

.. _InputPort_Creation:

Creating an InputPort
----------------------

An InputPort can be created by calling its constructor, but in general this is not necessary as a `Mechanism
<Mechanism>` can usually automatically create the InputPort(s) it needs when it is created.  For example, if the
Mechanism isbeing created within the `pathway <Process.pathway>` of a `Process`, its InputPort is created and assigned
as the `receiver <MappingProjection.receiver>` of a `MappingProjection` from the  preceding Mechanism in the `pathway
<Process.pathway>`.  InputPorts can also be specified in the **input_ports** argument of a Mechanism's constructor
(see `below <InputPort_Specification>`).

The `variable <InputPort.variable>` of an InputPort can be specified using the **variable** or **size** arguments of
its constructor.  It can also be specified using the **projections** argument, if neither **variable** nor **size** is
specified.  The **projections** argument is used to `specify Projections <Port_Projections>` to the InputPort. If
neither the **variable** nor **size** arguments is specified, then the value of the `Projections(s) <Projection>` or
their `sender <Projection_Base.sender>`\\s (all of which must be the same length) is used to determine the `variable
<InputPort.variable>` of the InputPort.

If an InputPort is created using its constructor, and a Mechanism is specified in the **owner** argument,
it is automatically assigned to that Mechanism.  Note that its `value <InputPort.value>` (generally determined
by the size of its `variable <InputPort.variable>` -- see `below <InputPort_Variable_and_Value>`) must
be compatible (in number and type of elements) with the item of its owner's `variable <Mechanism_Base.variable>` to
which it is assigned (see `below <InputPort_Variable_and_Value>` and `Mechanism <Mechanism_Variable_and_InputPorts>`).
If the **owner** argument is not specified, `initialization <Port_Deferred_Initialization>` is deferred.

.. _InputPort_Deferred_Initialization:

*Owner Assignment and Deferred Initialization*
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

An InputPort must be owned by a `Mechanism <Mechanism>`.  When InputPort is specified in the constructor for a
Mechanism (see `below <InputPort_Specification>`), it is automatically assigned to that Mechanism as its owner. If
the InputPort is created on its own, its `owner <Port.owner>` can specified in the **owner**  argument of its
constructor, in which case it is assigned to that Mechanism. If its **owner** argument is not specified, its
initialization is `deferred <Port_Deferred_Initialization>` until
COMMENT:
TBI: its `owner <Port_Base.owner>` attribute is assigned or
COMMENT
the InputPort is assigned to a Mechanism using the Mechanism's `add_ports <Mechanism_Base.add_ports>` method.

.. _InputPort_Primary:

*Primary InputPort*
~~~~~~~~~~~~~~~~~~~

Every Mechanism has at least one InputPort, referred to as its *primary InputPort*.  If InputPorts are not
`explicitly specified <InputPort_Specification>` for a Mechanism, a primary InputPort is automatically created
and assigned to its `input_port <Mechanism_Base.input_port>` attribute (note the singular), and also to the first
entry of the Mechanism's `input_ports <Mechanism_Base.input_ports>` attribute (note the plural).  The `value
<InputPort.value>` of the primary InputPort is assigned as the first (and often only) item of the Mechanism's
`variable <Mechanism_Base.variable>` and `input_values <Mechanism_Base.input_values>` attributes.

.. _InputPort_Specification:

*InputPort Specification*
~~~~~~~~~~~~~~~~~~~~~~~~~~

Specifying InputPorts when a Mechanism is created
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

InputPorts can be specified for a `Mechanism <Mechanism>` when it is created, in the **input_ports** argument of the
Mechanism's constructor (see `examples <Port_Constructor_Argument_Examples>` in Port), or in an *INPUT_PORTS* entry
of a parameter dictionary assigned to the constructor's **params** argument.  The latter takes precedence over the
former (that is, if an *INPUT_PORTS* entry is included in the parameter dictionary, any specified in the
**input_ports** argument are ignored).

    .. _InputPort_Replace_Default_Note:

    .. note::
       Assigning InputPorts to a Mechanism in its constructor **replaces** any that are automatically generated for
       that Mechanism (i.e., those that it creates for itself by default).  If any of those are needed, they must be
       explicitly specified in the list assigned to the **input_ports** argument, or the *INPUT_PORTS* entry of the
       parameter dictionary in the **params** argument.  The number of InputPorts specified must also be equal to
       the number of items in the Mechanism's `variable <Mechanism_Base.variable>` attribute.

.. _InputPort_Variable_and_Value:

*InputPort's* `variable <InputPort.variable>`, `value <InputPort.value>` *and Mechanism's* `variable <Mechanism_Base.variable>`
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

Each InputPort specified in the **input_ports** argument of a Mechanism's constructor must correspond to an item of
the Mechanism's `variable <Mechanism_Base.variable>` attribute (see `Mechanism <Mechanism_Variable_and_InputPorts>`),
and the `value <InputPort.value>` of the InputPort must be compatible with that item (that is, have the same number
and type of elements).  By default, this is also true of the InputPort's `variable <InputPort.variable>` attribute,
since the default `function <InputPort.function>` for an InputPort is a `LinearCombination`, the purpose of which
is to combine the inputs it receives and possibly modify the combined value (under the influence of any
`ModulatoryProjections <ModulatoryProjection>` it receives), but **not mutate its form**. Therefore, under most
circumstances, both the `variable <InputPort.variable>` of an InputPort and its `value <InputPort.value>` should
match the item of its owner's `variable <Mechanism_Base.variable>` to which the InputPort is assigned.

The format of an InputPort's `variable <InputPort.variable>` can be specified in a variety of ways.  The most
straightforward is in the **variable** argument of its constructor.  More commonly, however, it is determined by
the context in which it is being created, such as the specification for its owner Mechanism's `variable
<Mechanism_Base.variable>` or for the InputPort in the Mechanism's **input_ports** argument (see `below
<InputPort_Forms_of_Specification>` and `Mechanism InputPort specification <Mechanism_InputPort_Specification>`
for details).


Adding InputPorts to a Mechanism after it is created
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

InputPorts can also be **added** to a Mechanism, either by creating the InputPort on its own, and specifying the
Mechanism in the InputPort's **owner** argument, or by using the Mechanism's `add_ports <Mechanism_Base.add_ports>`
method (see `examples <Port_Create_Port_Examples>` in Port).

    .. _InputPort_Add_Port_Note:

    .. note::
       Adding InputPorts *does not replace* any that the Mechanism generates by default;  rather they are added to the
       Mechanism, and appended to the list of InputPorts in its `input_ports <Mechanism_Base.input_ports>` attribute.
       Importantly, the Mechanism's `variable <Mechanism_Base.variable>` attribute is extended with items that
       correspond to the `value <InputPort.value>` attribute of each added InputPort.  This may affect the
       relationship of the Mechanism's `variable <Mechanism_Base.variable>` to its `function
       <Mechanism_Base.function>`, as well as the number of its `OutputPorts <OutputPort>` (see `note
       <Mechanism_Add_InputPorts_Note>`).

If the name of an InputPort added to a Mechanism is the same as one that already exists, its name is suffixed with a
numerical index (incremented for each InputPort with that name; see `Registry_Naming`), and the InputPort is added to
the list (that is, it will *not* replace ones that already exist).

.. _InputPort_Forms_of_Specification:

Forms of Specification
^^^^^^^^^^^^^^^^^^^^^^

InputPorts can be specified in a variety of ways, that fall into three broad categories:  specifying an InputPort
directly; use of a `Port specification dictionary <Port_Specification>`; or by specifying one or more Components that
should project to the InputPort. Each of these is described below:

    .. _InputPort_Direct_Specification:

    **Direct Specification of an InputPort**

    * existing **InputPort object** or the name of one -- If this is used to specify an InputPort in the
      constructor for a Mechanism, its `value <InputPort.value>` must be compatible with the corresponding item of
      the owner Mechanism's `variable <Mechanism_Base.variable>` (see `Mechanism InputPort specification
      <Mechanism_InputPort_Specification>` and `InputPort_Compatability_and_Constraints` below).  If the InputPort
      belongs to another Mechanism, then an InputPort is created along with Projections(s) that `shadow the inputs
      <InputPort_Shadow_Inputs>` to the specified InputPort.
    ..
    * **InputPort class**, **keyword** *INPUT_PORT*, or a **string** -- this creates a default InputPort; if used
      to specify an InputPort in the constructor for a Mechanism, the item of the owner Mechanism's `variable
      <Mechanism_Base.variable>` to which the InputPort is assigned is used as the format for the InputPort`s
      `variable <InputPort.variable>`; otherwise, the default for the InputPort is used.  If a string is specified,
      it is used as the `name <InputPort.name>` of the InputPort (see `example <Port_Constructor_Argument_Examples>`).

    .. _InputPort_Specification_by_Value:

    * **value** -- this creates a default InputPort using the specified value as the InputPort's `variable
      <InputPort.variable>`; if used to specify an InputPort in the constructor for a Mechanism, the format must be
      compatible with the corresponding item of the owner Mechanism's `variable <Mechanism_Base.variable>` (see
      `Mechanism InputPort specification <Mechanism_InputPort_Specification>`, `example
      <port_value_Spec_Example>`, and discussion `below <InputPort_Compatability_and_Constraints>`).

    .. _InputPort_Specification_Dictionary:

    **InputPort Specification Dictionary**

    * **InputPort specification dictionary** -- this can be used to specify the attributes of an InputPort, using
      any of the entries that can be included in a `Port specification dictionary <Port_Specification>` (see
      `examples <Port_Specification_Dictionary_Examples>` in Port).  If the dictionary is used to specify an
      InputPort in the constructor for a Mechanism, and it includes a *VARIABLE* and/or *VALUE* or entry, the value
      must be compatible with the item of the owner Mechanism's `variable <Mechanism_Base.variable>` to which the
      InputPort is assigned (see `Mechanism InputPort specification <Mechanism_InputPort_Specification>`).

      The *PROJECTIONS* entry can include specifications for one or more Ports, Mechanisms and/or Projections that
      should project to the InputPort (including both `MappingProjections <MappingProjection>` and/or
      `ModulatoryProjections <ModulatoryProjection>`; however, this may be constrained by or have consequences for the
      InputPort's `variable <InputPort.variable>` (see `InputPort_Compatability_and_Constraints`).

      In addition to the standard entries of a `Port specification dictionary <Port_Specification>`, the dictionary
      can also include either or both of the following entries specific to InputPorts:

      * *WEIGHT*:<number>
          the value must be an integer or float, and is assigned as the value of the InputPort's `weight
          <InputPort.weight>` attribute (see `weight and exponent <InputPort_Weights_And_Exponents>`);
          this takes precedence over any specification in the **weight** argument of the InputPort's constructor.

      * *EXPONENT*:<number>
          the value must be an integer or float, and is assigned as the value of the InputPort's `exponent
          <InputPort.exponent>` attribute (see `weight and exponent <InputPort_Weights_And_Exponents>`);
          this takes precedence over any specification in the **exponent** argument of the InputPort's constructor.

    .. _InputPort_Projection_Source_Specification:

    **Specification of an InputPort by Components that Project to It**

    COMMENT:
    `examples
      <Port_Projections_Examples>` in Port)
    COMMENT

    COMMENT:
    ?? PUT IN ITS OWN SECTION ABOVE OR BELOW??
    Projections to an InputPort can be specified either as attributes, in the constructor for an
    InputPort (in its **projections** argument or in the *PROJECTIONS* entry of an `InputPort specification dictionary
    <InputPort_Specification_Dictionary>`), or used to specify the InputPort itself (using one of the
    `InputPort_Forms_of_Specification` described above. See `Port Projections <Port_Projections>` for additional
    details concerning the specification of
    Projections when creating a Port.
    COMMENT

    An InputPort can also be specified by specifying one or more Ports, Mechanisms or Projections that should project
    to it, as described below.  Specifying an InputPort in this way creates both the InputPort and any of the
    specified or implied Projection(s) to it (if they don't already exist). `MappingProjections <MappingProjection>`
    are assigned to the InputPort's `path_afferents <Port.path_afferents>` attribute, while `ControlProjections
    <ControlProjection>` and `GatingProjections <GatingProjection>` to its `mod_afferents <Port.mod_afferents>`
    attribute. Any of the following can be used to specify an InputPort by the Components that projection to it (see
    `below <InputPort_Compatability_and_Constraints>` for an explanation of the relationship between the `value` of
    these Components and the InputPort's `variable <InputPort.variable>`):

    * **OutputPort, GatingSignal, Mechanism, or list with any of these** -- creates an InputPort with Projection(s)
      to it from the specified Port(s) or Mechanism(s).  For each Mechanism specified, its `primary OutputPort
      <OutputPort_Primary>` (or GatingSignal) is used.
    ..
    * **Projection** -- any form of `Projection specification <Projection_Specification>` can be
      used;  creates an InputPort and assigns it as the Projection's `receiver <Projection_Base.receiver>`.

    .. _InputPort_Tuple_Specification:

    * **InputPort specification tuples** -- these are convenience formats that can be used to compactly specify an
      InputPort and Projections to it any of the following ways:

        .. _InputPort_Port_Mechanism_Tuple:

        * **2-item tuple:** *(<Port name or list of Port names>, <Mechanism>)* -- 1st item must be the name of an
          `OutputPort` or `ModulatorySignal`, or a list of such names, and the 2nd item must be the Mechanism to
          which they all belong.  Projections of the relevant types are created for each of the specified Ports
          (see `Port 2-item tuple <Port_2_Item_Tuple>` for additional details).

        * **2-item tuple:** *(<value, Port specification, or list of Port specs>, <Projection specification>)* --
          this is a contracted form of the 4-item tuple described below;

        * **3 or 4-item tuple:** *(<value, Port spec, or list of Port specs>, weight, exponent, Projection
          specification)* -- this allows the specification of Port(s) that should project to the InputPort, together
          with a specification of the InputPort's `weight <InputPort.weight>` and/or `exponent <InputPort.exponent>`
          attributes of the InputPort, and (optionally) the Projection(s) to it.  This can be used to compactly
          specify a set of Ports that project the InputPort, while using the 4th item to determine its variable
          (e.g., using the matrix of the Projection specification) and/or attributes of the Projection(s) to it. Each
          tuple must have at least the following first three items (in the order listed), and can include the fourth:

            * **value, Port specification, or list of Port specifications** -- specifies either the `variable
              <InputPort.variable>` of the InputPort, or one or more Ports that should project to it.  The Port
              specification(s) can be a (Port name, Mechanism) tuple (see above), and/or include Mechanisms (in which
              case their `primary OutputPort <OutputPortPrimary>` is used.  All of the Port specifications must be
              consistent with (that is, their `value <Port_Base.value>` must be compatible with the `variable
              <Projection_Base.variable>` of) the Projection specified in the fourth item if that is included;

            * **weight** -- must be an integer or a float; multiplies the `value <InputPort.value>` of the InputPort
              before it is combined with others by the Mechanism's `function <Mechanism.function>` (see
              ObjectiveMechanism for `examples <ObjectiveMechanism_Weights_and_Exponents_Example>`);

            * **exponent** -- must be an integer or float; exponentiates the `value <InputPort.value>` of the
              InputPort before it is combined with others by the ObjectiveMechanism's `function
              <ObjectiveMechanism.function>` (see ObjectiveMechanism for `examples
              <ObjectiveMechanism_Weights_and_Exponents_Example>`);

            * **Projection specification** (optional) -- `specifies a Projection <Projection_Specification>` that
              must be compatible with the Port specification(s) in the 1st item; if there is more than one Port
              specified, and the Projection specification is used, all of the Ports
              must be of the same type (i.e.,either OutputPorts or GatingSignals), and the `Projection
              Specification <Projection_Specification>` cannot be an instantiated Projection (since a
              Projection cannot be assigned more than one `sender <Projection_Base.sender>`).

    .. _InputPort_Shadow_Inputs:

    * **InputPorts of Mechanisms to shadow** -- either of the following can be used to create InputPorts that
      receive the same inputs as ("shadow") the ones specified:

      * *InputPort or [InputPort, ...]* -- each InputPort must belong to an existing Mechanism; creates a new
        InputPort for each one specified, along with Projections to it that parallel those of the one specified
        (see below).

      * *{SHADOW_INPUTS: <InputPort or Mechanism or [<InputPort or Mechanism>,...]>}* -- any InputPorts specified
        must belong to an existing Mechanism;  creates a new InputPort for each one specified, and for each of the
        InputPorts belonging to any Mechanisms specified, along with Projections to them that parallel those of the
        one(s) specified (see below).

      For each InputPort specified, and all of the InputPorts belonging to any Mechanisms specified, a new InputPort
      is created along with Projections to it that parallel those received by the corresponding InputPort in the
      list.  In other words, for each InputPort specified, a new one is created that receives exactly the same inputs
      from the same `senders  <Projection_Base.sender>` as the ones specified.

.. _InputPort_Compatability_and_Constraints:

InputPort `variable <InputPort.variable>`: Compatibility and Constraints
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The `variable <InputPort.variable>` of an InputPort must be compatible with the item of its owner Mechanism's
`variable <Mechanism_Base.variable>` to which it is assigned (see `Mechanism_Variable_and_InputPorts`). This may
have consequences that must be taken into account when `specifying an InputPort by Components that project to it
<InputPort_Projection_Source_Specification>`.  These depend on the context in which the specification is made, and
possibly the value of other specifications.  These considerations and how they are handled are described below,
starting with constraints that are given the highest precedence:

  *  **InputPort is** `specified in a Mechanism's constructor <Mechanism_InputPort_Specification>` and the
    **default_variable** argument for the Mechanism is also specified -- the item of the variable to which the
    `InputPort is assigned <Mechanism_Variable_and_InputPorts>` is used to determine the InputPort's `variable must
    <InputPort.variable>`.  Any other specifications of the InputPort relevant to its `variable <InputPort.variable>`
    must be compatible with this (for example, `specifying it by value <InputPort_Specification_by_Value>` or by a
    `MappingProjection` or `OutputPort` that projects to it (see `above <InputPort_Projection_Source_Specification>`).

    COMMENT:
    ***XXX EXAMPLE HERE
    COMMENT
  ..
  * **InputPort is specified on its own**, or the **default_variable** argument of its Mechanism's constructor
    is not specified -- any direct specification of the InputPort's `variable <InputPort.variable>` is used to
    determine its format (e.g., `specifying it by value <InputPort_Specification_by_Value>`, or a *VARIABLE* entry
    in an `InputPort specification dictionary <InputPort_Specification_Dictionary>`.  In this case, the value of any
    `Components used to specify the InputPort <InputPort_Projection_Source_Specification>` that are relevant to its
    `variable <InputPort.variable>` must be compatible with it (see below).

    COMMENT:
    ***XXX EXAMPLE HERE
    COMMENT
  ..
  * If the InputPort's `variable <InputPort.variable>` is not constrained by any of the conditions above,
    then its format is determined by the `specification of Components that project to it
    <InputPort_Projection_Source_Specification>`:

    * **More than one Component is specified with the same `value` format** -- that format is used to determine
      the format of the InputPort's `variable <InputPort.variable>`.

    * **More than one Component is specified with different `value` formats** -- the InputPort's `variable
      <InputPort.variable>` is determined by item of the default `variable <Mechanism_Base.variable>` for
      the class of its owner Mechanism.

    * **A single Component is specified** -- its `value` is used to determine the format of the InputPort's
      `variable <InputPort.variable>`;  if the Component is a(n):

      * **MappingProjection** -- can be specified by its class, an existing MappingProjection, or a matrix:

        * `MappingProjection` **class** -- a default value is used both the for the InputPort's `variable
          <InputPort.variable>` and the Projection's `value <Projection_Base.value>` (since the Projection's
          `sender <Projection_Base.sender>` is unspecified, its `initialization is deferred
          <Projection_Deferred_Initialization>`.

        * **Existing MappingProjection** -- then its `value <Projection_Base.value>` determines the
          InputPort's `variable <InputPort.variable>`.

        * `Matrix specification <MappingProjection_Matrix_Specification>` -- its receiver dimensionality determines the
          format of the InputPort's `variable <InputPort.variable>`. For a standard 2d "weight" matrix (i.e., one that
          maps a 1d array from its `sender <Projection_Base.sender>` to a 1d array of its `receiver
          <Projection_Base.receiver>`), the receiver dimensionality is its outer dimension (axis 1, or its number of
          columns).  However, if the `sender <Projection_Base.sender>` has more than one dimension, then the
          dimensionality of the receiver (used for the InputPort's `variable <InputPort.variable>`) is the
          dimensionality of the matrix minus the dimensionality of the sender's `value <OutputPort.value>`
          (see `matrix dimensionality <Mapping_Matrix_Dimensionality>`).

      * **OutputPort or ProcessingMechanism** -- the `value <OutputPort.value>` of the OutputPort (if it is a
        Mechanism, then its `primary OutputPort <OutputPort_Primary>`) determines the format of the InputPort's
        `variable <InputPort.variable>`, and a MappingProjection is created from the OutputPort to the InputPort
        using an `IDENTITY_MATRIX`.  If the InputPort's `variable <InputPort.variable>` is constrained (as in some
        of the cases above), then a `FULL_CONNECTIVITY_MATRIX` is used which maps the shape of the OutputPort's `value
        <OutputPort.value>` to that of the InputPort's `variable <InputPort.variable>`.

      * **GatingProjection, GatingSignal or GatingMechanism** -- any of these can be used to specify an InputPort;
        their `value` does not need to be compatible with the InputPort's `variable <InputPort.variable>`, however
        it does have to be compatible with the `modulatory parameter <Function_Modulatory_Params>` of the InputPort's
        `function <InputPort.function>`.

.. _InputPort_Structure:

Structure
---------

Every InputPort is owned by a `Mechanism <Mechanism>`. It can receive one or more `MappingProjections
<MappingProjection>` from other Mechanisms, as well as from the Process or System to which its owner belongs (if it
is the `ORIGIN` Mechanism for that Process or System).  It has the following attributes, that includes ones specific
to, and that can be used to customize the InputPort:

* `projections <Port.projections>` -- all of the `Projections <Projection>` received by the InputPort.

.. _InputPort_Afferent_Projections:

* `path_afferents <Port.path_afferents>` -- `MappingProjections <MappingProjection>` that project to the InputPort,
  the `value <Projection_Base.value>`\\s of which are combined by the InputPort's `function <InputPort.function>`,
  possibly modified by its `mod_afferents <InputPort_mod_afferents>`, and assigned to the corresponding item of the
  owner Mechanism's `variable <Mechanism_Base.variable>`.

* `mod_afferents <InputPort_mod_afferents>` -- `GatingProjections <GatingProjection>` that project to the InputPort,
  the `value <GatingProjection.value>` of which can modify the InputPort's `value <InputPort.value>` (see the
  descriptions of Modulation under `ModulatorySignals <ModulatorySignal_Modulation>` and `GatingSignals
  <GatingSignal_Modulation>` for additional details).  If the InputPort receives more than one GatingProjection,
  their values are combined before they are used to modify the `value <InputPort.value>` of InputPort.

.. _InputPort_Variable:

* `variable <InputPort.variable>` -- serves as the template for the `value <Projection_Base.value>` of the
  `Projections <Projection>` received by the InputPort:  each must be compatible with (that is, match both the
  number and type of elements of) the InputPort's `variable <InputPort.variable>` (see `Mapping_Matrix` for additonal
  details). In general, this must also be compatible with the item of the owner Mechanism's `variable
  <Mechanism_Base.variable>` to which the InputPort is assigned (see `above <InputPort_Variable_and_Value>` and
  `Mechanism InputPort specification <Mechanism_InputPort_Specification>`).

.. _InputPort_Function:

* `function <InputPort.function>` -- combines the `value <Projection_Base.value>` of all of the
  `Projections <Projection>` received by the InputPort, and assigns the result to the InputPort's `value
  <InputPort.value>` attribute.  The default function is `LinearCombination` that performs an elementwise (Hadamard)
  sums the values. However, the parameters of the `function <InputPort.function>` -- and thus the `value
  <InputPort.value>` of the InputPort -- can be modified by any `GatingProjections <GatingProjection>` received by
  the InputPort (listed in its `mod_afferents <Port.mod_afferents>` attribute.  A custom function can also be
  specified, so long as it generates a result that is compatible with the item of the Mechanism's `variable
  <Mechanism_Base.variable>` to which the `InputPort is assigned <Mechanism_InputPorts>`.

.. _InputPort_Value:

* `value <InputPort.value>` -- the result returned by its `function <InputPort.function>`,
  after aggregating the value of the `PathProjections <PathwayProjection>` it receives, possibly modified by any
  `GatingProjections <GatingProjection>` received by the InputPort. It must be compatible with the
  item of the owner Mechanism's `variable <Mechanism_Base.variable>` to which the `InputPort has been assigned
  <Mechanism_InputPorts>` (see `above <InputPort_Variable_and_Value>` and `Mechanism InputPort specification
  <Mechanism_InputPort_Specification>`).

.. _InputPort_Weights_And_Exponents:

* `weight <InputPort.weight>` and `exponent <InputPort.exponent>` -- these can be used by the Mechanism to which the
  InputPort belongs when that combines the `value <InputPort.value>`\\s of its Ports (e.g., an ObjectiveMechanism
  uses the weights and exponents assigned to its InputPorts to determine how the values it monitors are combined by
  its `function <ObjectiveMechanism>`).  The value of each must be an integer or float, and the default is 1 for both.

.. _InputPort_Execution:

Execution
---------

An InputPort cannot be executed directly.  It is executed when the Mechanism to which it belongs is executed.
When this occurs, the InputPort executes any `Projections <Projection>` it receives, calls its `function
<InputPort.function>` to combines the values received from any `MappingProjections <MappingProjection>` it receives
(listed in its its `path_afferents  <Port.path_afferents>` attribute) and modulate them in response to any
`GatingProjections <GatingProjection>` (listed in its `mod_afferents <Port.mod_afferents>` attribute),
and then assigns the result to the InputPort's `value <InputPort.value>` attribute. This, in turn, is assigned to
the item of the Mechanism's `variable <Mechanism_Base.variable>` and `input_values <Mechanism_Base.input_values>`
attributes  corresponding to that InputPort (see `Mechanism Variable and InputPorts
<Mechanism_Variable_and_InputPorts>` for additional details).

.. _InputPort_Class_Reference:

Class Reference
---------------

"""
import inspect
import numbers
import warnings

import collections
import numpy as np
import typecheck as tc

from psyneulink.core.components.component import DefaultsFlexibility
from psyneulink.core.components.functions.function import Function
from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.functions.combinationfunctions import CombinationFunction, LinearCombination, Reduce
from psyneulink.core.components.functions.statefulfunctions.memoryfunctions import Buffer
from psyneulink.core.components.ports.outputport import OutputPort
from psyneulink.core.components.ports.port import PortError, Port_Base, _instantiate_port_list, port_type_keywords
from psyneulink.core.globals.context import ContextFlags, handle_external_context
from psyneulink.core.globals.keywords import \
    COMBINE, CONTROL_SIGNAL, EXPONENT, FUNCTION, GATING_SIGNAL, INPUT_PORT, INPUT_PORTS, INPUT_PORT_PARAMS, \
    LEARNING_SIGNAL, MAPPING_PROJECTION, MATRIX, NAME, OPERATION, OUTPUT_PORT, OUTPUT_PORTS, OWNER,\
    PARAMS, PRODUCT, PROJECTIONS, REFERENCE_VALUE, \
    SENDER, SHADOW_INPUTS, SHADOW_INPUT_NAME, SIZE, PORT_TYPE, SUM, VALUE, VARIABLE, WEIGHT
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.basepreferenceset import is_pref_set
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel
from psyneulink.core.globals.utilities import append_type_to_name, convert_to_np_array, is_numeric, iscompatible, kwCompatibilityLength

__all__ = [
    'InputPort', 'InputPortError', 'port_type_keywords', 'SHADOW_INPUTS',
]

port_type_keywords = port_type_keywords.update({INPUT_PORT})

# InputPortPreferenceSet = BasePreferenceSet(log_pref=logPrefTypeDefault,
#                                                          reportOutput_pref=reportOutputPrefTypeDefault,
#                                                          verbose_pref=verbosePrefTypeDefault,
#                                                          param_validation_pref=paramValidationTypeDefault,
#                                                          level=PreferenceLevel.TYPE,
#                                                          name='InputPortClassPreferenceSet')

WEIGHT_INDEX = 1
EXPONENT_INDEX = 2

DEFER_VARIABLE_SPEC_TO_MECH_MSG = "InputPort variable not yet defined, defer to Mechanism"

class InputPortError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class InputPort(Port_Base):
    """
    InputPort(                                     \
        variable=None,                             \
        reference_value=None,                      \
        function=LinearCombination(operation=SUM), \
        combine=None,                              \
        projections=None,                          \
        weight=None,                               \
        exponent=None,                             \
        internal_only=False)

    Subclass of `Port <Port>` that calculates and represents the input to a `Mechanism <Mechanism>` from one or more
    `PathwayProjections <PathwayProjection>`.  See `Port_Class_Reference` for additional arguments and attributes.

    COMMENT:

    PortRegistry
    -------------
        All InputPorts are registered in PortRegistry, which maintains an entry for the subclass,
        a count for all instances of it, and a dictionary of those instances

    COMMENT

    Arguments
    ---------

    reference_value : number, list or np.ndarray
        the value of the item of the owner Mechanism's `variable <Mechanism_Base.variable>` attribute to which
        the InputPort is assigned; used as the template for the InputPort's `value <InputPort.value>` attribute.

    variable : number, list or np.ndarray
        specifies the shape of the  InputPort's `variable <InputPort.variable>`, which may be used to define the
        shape of the `matrix <MappingProjection.matrix>` parameter of the `MappingProjection` that projects to the
        Inputport (see `InputPort_Variable` for additional details).

    function : Function or method : default LinearCombination(operation=SUM)
        specifies the function applied to the variable. The default value combines the `values
        <Projection_Base.value>` of the `Projections <Projection>` received by the InputPort.  Any function
        can be assigned, however:  a) it must produce a result that has the same format (number and type of elements)
        as the item of its owner Mechanism's `variable <Mechanism_Base.variable>` to which the InputPort has been
        assigned;  b) if it is not a CombinationFunction, it may produce unpredictable results if the InputPort
        receives more than one Projection (see `function <InputPort.function>`.

    combine : SUM or PRODUCT : default None
        specifies the **operation** argument used by the default `LinearCombination` function, which determines how the
        `value <Projection_Base.value>` of the InputPort's `projections <Port.projections>` are combined.  This is a
        convenience argument, that allows the **operation** to be specified without having to specify the
        LinearCombination function; it assumes that LinearCombination (the default) is used as the InputPort's function
        -- if it conflicts with a specification of **function** an error is generated.

    projections : list of Projection specifications
        specifies the `MappingProjection(s) <MappingProjection>`, `ControlProjection(s) <ControlProjection>` and/or
        `GatingProjection(s) <GatingProjection>` to be received by the InputPort, and that are listed in its
        `path_afferents <Port.path_afferents>` and `mod_afferents <Port.mod_afferents>` attributes,
        respectively (see `InputPort_Compatability_and_Constraints` for additional details).  If **projections** but
        neither **variable** nor **size** are specified, then the `value <Projection_Base.value>` of the Projection(s)
        or their `senders <Projection_Base.sender>` specified in **projections** argument are used to determine the
        InputPort's `variable <InputPort.variable>`.

    weight : number : default 1
        specifies the value of the `weight <InputPort.weight>` attribute of the InputPort.

    exponent : number : default 1
        specifies the value of the `exponent <InputPort.exponent>` attribute of the InputPort.

    internal_only : bool : False
        specifies whether the InputPort requires external input when its `owner <Port.owner>` is the `INPUT`
        `Node <Composition_Nodes>` of a `Composition (see `internal_only <InputPort.internal_only>` for details).

    Attributes
    ----------

    variable : value, list or np.ndarray
        the template for the `value <Projection_Base.value>` of each Projection that the InputPort receives,
        each of which must match the format (number and types of elements) of the InputPort's
        `variable <InputPort.variable>`.  If neither the **variable** or **size** argument is specified, and
        **projections** is specified, then `variable <InputPort.variable>` is assigned the `value
        <Projection_Base.value>` of the Projection(s) or its `sender <Projection_Base.sender>`.

    function : Function
        If it is a `CombinationFunction`, it combines the `values <Projection_Base.value>` of the `PathwayProjections
        <PathwayProjection>` (e.g., `MappingProjections <MappingProjection>`) received by the InputPort  (listed in
        its `path_afferents <Port.path_afferents>` attribute), under the possible influence of `GatingProjections
        <GatingProjection>` received by the InputPort (listed in its `mod_afferents <Port.mod_afferents>` attribute).
        The result is assigned to the InputPort's `value <InputPort.value>` attribute. For example, the default
        (`LinearCombination` with *SUM* as it **operation**) performs an element-wise (Hadamard) sum of its Projection
        `values <Projection_Base.value>`, and assigns to `value <InputPort.value>` an array that is of the same length
        as each of the Projection `values <Projection_Base.value>`.  If the InputPort receives only one Projection,
        then any other function can be applied and it will generate a value that is the same length as the Projection's
        `value <Projection_Base.value>`. However, if the InputPort receives more than one Projection and uses a function
        other than a CombinationFunction, a warning is generated and only the `value <Projection_Base.value>` of the
        first Projection list in `path_afferents <Port.path_afferents>` is used by the function, which may generate
        unexpected results when executing the Mechanism or Composition to which it belongs.

    value : value or ndarray
        the output of the InputPort's `function <InputPort.function>`, that is assigned to an item of the owner
        Mechanism's `variable <Mechanism_Base.variable>` attribute.

    label : string or number
        the string label that represents the current `value <InputPort.value>` of the InputPort, according to the
        owner mechanism's `input_labels_dict <Mechanism.input_labels_dict>`. If the current `value <InputPort.value>`
        of the InputPort does not have a corresponding label, then the numeric `value <InputPort.value>` is returned.

    weight : number
        see `weight and exponent <InputPort_Weights_And_Exponents>` for description.

    exponent : number
        see `weight and exponent <InputPort_Weights_And_Exponents>` for description.

    internal_only : bool
        determines whether `input from a Composition <Composition_Execution_Input>` must be specified for this
        InputPort from a Composition's `execution method <Composition_Execution_Method>` if the InputPort's `owner
        <Port.owner>` is an `INPUT` `Node <Composition_Nodes>` of that Composition; if `True`, external input is
        *not* required or allowed.

    name : str
        the name of the InputPort; if it is not specified in the **name** argument of the constructor, a default is
        assigned by the InputPortRegistry of the Mechanism to which the InputPort belongs.  Note that some Mechanisms
        automatically create one or more non-default InputPorts, that have pre-specified names.  However, if any
        InputPorts are specified in the **input_ports** argument of the Mechanism's constructor, those replace those
        InputPorts (see `note <Mechanism_Default_Port_Suppression_Note>`), and `standard naming conventions
        <Registry_Naming>` apply to the InputPorts specified, as well as any that are added to the Mechanism once it
        is created (see `note <Port_Naming_Note>`).

    """

    #region CLASS ATTRIBUTES

    componentType = INPUT_PORT
    paramsType = INPUT_PORT_PARAMS

    portAttributes = Port_Base.portAttributes | {WEIGHT, EXPONENT}

    connectsWith = [OUTPUT_PORT,
                    LEARNING_SIGNAL,
                    GATING_SIGNAL,
                    CONTROL_SIGNAL
                    ]
    connectsWithAttribute = [OUTPUT_PORTS]
    projectionSocket = SENDER
    modulators = [GATING_SIGNAL, CONTROL_SIGNAL]
    canReceive = modulators + [MAPPING_PROJECTION]
    projection_type = MAPPING_PROJECTION

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TYPE_DEFAULT_PREFERENCES
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     PREFERENCE_SET_NAME: 'InputPortCustomClassPreferences',
    #     PREFERENCE_KEYWORD<pref>: <setting>...}

    # Note: the following enforce encoding as 1D np.ndarrays (one variable/value array per port)
    variableEncodingDim = 1
    valueEncodingDim = 1

    class Parameters(Port_Base.Parameters):
        """
            Attributes
            ----------

                combine
                    see `combine <InputPort.combine>`

                    :default value: None
                    :type:

                exponent
                    see `exponent <InputPort.exponent>`

                    :default value: None
                    :type:

                function
                    see `function <InputPort_Function>`

                    :default value: `LinearCombination`
                    :type: `Function`

                internal_only
                    see `internal_only <InputPort.internal_only>`

                    :default value: False
                    :type: ``bool``

                shadow_inputs
                    specifies whether InputPort shadows inputs of another InputPort;
                    if not None, must be assigned another InputPort

                    :default value: None
                    :type:
                    :read only: True

                weight
                    see `weight <InputPort.weight>`

                    :default value: None
                    :type:
        """
        function = Parameter(LinearCombination(operation=SUM), stateful=False, loggable=False)
        weight = Parameter(None, modulable=True)
        exponent = Parameter(None, modulable=True)
        combine = None
        internal_only = Parameter(False, stateful=False, loggable=False, pnl_internal=True)
        shadow_inputs = Parameter(None, stateful=False, loggable=False, read_only=True, pnl_internal=True, structural=True)

    #endregion

    @handle_external_context()
    @tc.typecheck
    def __init__(self,
                 owner=None,
                 reference_value=None,
                 variable=None,
                 size=None,
                 function=None,
                 projections=None,
                 combine:tc.optional(tc.enum(SUM,PRODUCT))=None,
                 weight=None,
                 exponent=None,
                 internal_only: tc.optional(bool) = None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 context=None,
                 **kwargs):

        if variable is None and size is None and projections is not None:
            variable = self._assign_variable_from_projection(variable, size, projections)

        # If combine argument is specified, save it along with any user-specified function for _validate_params()
        if combine:
            self.combine_function_args = (combine, function)

        # If owner or reference_value has not been assigned, defer init to Port._instantiate_projection()
        # if owner is None or (variable is None and reference_value is None and projections is None):
        if owner is None:
            # Temporarily name InputPort
            self._assign_deferred_init_name(name)
            # Store args for deferred initialization
            self._store_deferred_init_args(**locals())

            # Flag for deferred initialization
            self.initialization_status = ContextFlags.DEFERRED_INIT
            return

        self.reference_value = reference_value

        # Validate sender (as variable) and params, and assign to variable
        # Note: pass name of owner (to override assignment of componentName in super.__init__)
        super(InputPort, self).__init__(
            owner,
            variable=variable,
            size=size,
            projections=projections,
            function=function,
            weight=weight,
            exponent=exponent,
            internal_only=internal_only,
            shadow_inputs=None,
            params=params,
            name=name,
            prefs=prefs,
            context=context,
        )

        if self.name is self.componentName or self.componentName + '-' in self.name:
            self._assign_default_port_Name()

    def _assign_variable_from_projection(self, variable, size, projections):
        """Assign variable to value of Projection in projections
        """
        from psyneulink.core.components.projections.projection import \
            Projection, _parse_connection_specs

        if not isinstance(projections, list):
            projections = [projections]

        # Use only first specification in the list returned, and assume any others are the same size
        #     (which they must be); leave validation of this to _instantiate_projections_to_port
        proj_spec = _parse_connection_specs(InputPort, self, projections)[0]

        if isinstance(proj_spec.projection, Projection):
            variable = proj_spec.projection.defaults.value
        elif isinstance(proj_spec.port, OutputPort):
            variable = proj_spec.port.defaults.value
        else:
            raise InputPortError(f"Unrecognized specification for \'{PROJECTIONS}\' arg of {self.name}.")

        return variable

    def _validate_params(self, request_set, target_set=None, context=None):
        """Validate weights and exponents

        This needs to be done here
            (so that they can be ignored if not specified here or in the function)
        """

        super()._validate_params(request_set=request_set, target_set=target_set, context=context)

        # Make sure **combine** and **function** args specified in constructor don't conflict
        if hasattr(self, 'combine_function_args'):
            combine, function = self.combine_function_args
            if function:
                owner_name = ""
                if self.owner:
                    owner_name = f" for InputPort of {self.owner.name}."
                if isinstance(function, LinearCombination):
                    # specification of combine conflicts with operation specified for LinearCombination in function arg
                    if function.operation != combine:
                        raise InputPortError(f"Specification of {repr(COMBINE)} argument ({combine.upper()}) "
                                             f"conflicts with specification of {repr(OPERATION)} "
                                             f"({function.operation.upper()}) for LinearCombination in "
                                             f"{repr(FUNCTION)} argument{owner_name}.")
                    else:
                        # LinearFunction has been specified with same operation as specified for combine,
                        # so delete combine_function_args attribute so it is not seen in _instantiate_function
                        # in order to leave function intact (as it may have other parameters specified by user)
                        del self.combine_function_args
                # combine assumes LinearCombination, but Function other than LinearCombination specified for function
                elif isinstance(function, Function):
                    raise InputPortError(f"Specification of {repr(COMBINE)} argument ({combine.upper()}) "
                                         f"conflicts with Function specified in {repr(FUNCTION)} argument "
                                         f"({function.name}){owner_name}.")
                # combine assumes LinearCombination, but class other than LinearCombination specified for function
                elif isinstance(function, type):
                    if not issubclass(function, LinearCombination):
                        raise InputPortError(f"Specification of {repr(COMBINE)} argument ({combine.upper()}) "
                                             f"conflicts with Function specified in {repr(FUNCTION)} argument "
                                             f"({function.__name__}){owner_name}.")
                else:
                    raise InputPortError(f"PROGRAM ERROR: unrecognized specification for function argument "
                                         f"({function}){owner_name}.")

        if WEIGHT in target_set and target_set[WEIGHT] is not None:
            if not isinstance(target_set[WEIGHT], (int, float)):
                raise InputPortError(f"'{WEIGHT}' parameter of {self.name} for {self.owner.name} "
                                     f"({target_set[WEIGHT]}) must be an int or float.")

        if EXPONENT in target_set and target_set[EXPONENT] is not None:
            if not isinstance(target_set[EXPONENT], (int, float)):
                raise InputPortError(f"'{EXPONENT}' parameter of {self.name} for {self.owner.name}"
                                     f"({ target_set[EXPONENT]}) must be an int or float.")

    def _validate_against_reference_value(self, reference_value):
        """Validate that Port.value is compatible with reference_value

        reference_value is the item of the owner Mechanism's variable to which the InputPort is assigned
        """
        match_len_option = {kwCompatibilityLength:False}
        if reference_value is not None and not iscompatible(reference_value, self.defaults.value, **match_len_option):
            name = self.name or ""
            raise InputPortError(f"Value specified for {name} {self.componentName} of {self.owner.name} "
                                 f"({self.defaults.value}) is not compatible with its expected format "
                                 f"({reference_value}).")

    def _instantiate_function(self, function, function_params=None, context=None):
        """If combine option was specified in constructor, assign as operation argument of LinearCombination function"""
        if hasattr(self, 'combine_function_args'):
            function = LinearCombination(operation=self.combine_function_args[0])
            del self.combine_function_args
        super()._instantiate_function(function=function, context=context)
        self._use_1d_variable = False
        if not isinstance(self.function, CombinationFunction):
            self._use_1d_variable = True
            self.function._variable_shape_flexibility = DefaultsFlexibility.RIGID
        else:
            self.function._variable_shape_flexibility = DefaultsFlexibility.FLEXIBLE

    def _instantiate_projections(self, projections, context=None):
        """Instantiate Projections specified in PROJECTIONS entry of params arg of Port's constructor

        Call _instantiate_projections_to_port to assign:
            PathwayProjections to .path_afferents
            ModulatoryProjections to .mod_afferents
        """
        self._instantiate_projections_to_port(projections=projections, context=context)

    def _check_for_duplicate_projections(self, projection):
        """Check if projection is redundant with one in path_afferents of InputPort

        Check for any instantiated projection in path_afferents with the same sender as projection
        or one in deferred_init status with sender specification that is the same type as projection.

        Returns redundant Projection if found, otherwise False.
        """

        try:
            self.path_afferents
        except:
            if self.initialization_status == ContextFlags.DEFERRED_INIT:
                raise InputPortError(f"Attempt to assign Projection ('{projection}') "
                                     f"to InputPort ('{self.name}') that is in deferred init")
            else:
                raise InputPortError(f"No 'path_afferents' for {self.name}")

        # FIX: 7/22/19 - CHECK IF SENDER IS SPECIFIED AS MECHANISM AND, IF SO, CHECK ITS PRIMARY_OUTPUT_PORT
        duplicate = next(iter([proj for proj in self.path_afferents
                               if ((proj.sender == projection.sender and proj != projection)
                                   or (proj.initialization_status == ContextFlags.DEFERRED_INIT
                                       and proj._init_args[SENDER] == type(projection.sender)))]), None)
        if duplicate and self.verbosePref or self.owner.verbosePref:
            from psyneulink.core.components.projections.projection import Projection
            warnings.warn(f'{Projection.__name__} from {projection.sender.name}  {projection.sender.__class__.__name__}'
                          f' of {projection.sender.owner.name} to {self.name} {self.__class__.__name__} of '
                          f'{self.owner.name} already exists; will ignore additional one specified ({projection.name}).')
        return duplicate

    def _parse_function_variable(self, variable, context=None):
        variable = super()._parse_function_variable(variable, context)
        try:
            if self._use_1d_variable and variable.ndim > 1:
                return np.array(variable[0])
        except AttributeError:
            pass
        return variable

    def _get_variable_from_projections(self, context=None):
        """
            Call self.function with self._path_proj_values

            If variable is None there are no active PathwayProjections in the Composition being run,
            return None so that it is ignored in execute method (i.e., not combined with base_value)
        """
        # Check for Projections that are active in the Composition being run
        path_proj_values = [
            proj.parameters.value._get(context)
            for proj in self.path_afferents
            if self.afferents_info[proj].is_active_in_composition(context.composition)
        ]

        if len(path_proj_values) > 0:
            return convert_to_np_array(path_proj_values)
        else:
            return None

    def _get_primary_port(self, mechanism):
        return mechanism.input_port

    @tc.typecheck
    def _parse_port_specific_specs(self, owner, port_dict, port_specific_spec):
        """Get weights, exponents and/or any connections specified in an InputPort specification tuple

        Tuple specification can be:
            (port_spec, connections)
            (port_spec, weights, exponents, connections)

        See Port._parse_port_specific_spec for additional info.
.
        Returns:
             - port_spec:  1st item of tuple if it is a numeric value;  otherwise None
             - params dict with WEIGHT, EXPONENT and/or PROJECTIONS entries if any of these was specified.

        """
        # FIX: ADD FACILITY TO SPECIFY WEIGHTS AND/OR EXPONENTS FOR INDIVIDUAL OutputPort SPECS
        #      CHANGE EXPECTATION OF *PROJECTIONS* ENTRY TO BE A SET OF TUPLES WITH THE WEIGHT AND EXPONENT FOR IT
        #      THESE CAN BE USED BY THE InputPort's LinearCombination Function
        #          (AKIN TO HOW THE MECHANISM'S FUNCTION COMBINES InputPort VALUES)
        #      THIS WOULD ALLOW AN ADDITIONAL HIERARCHICAL LEVEL FOR NESTING ALGEBRAIC COMBINATION OF INPUT VALUES
        #      TO A MECHANISM
        from psyneulink.core.components.projections.projection import Projection, _parse_connection_specs

        params_dict = {}
        port_spec = port_specific_spec

        if isinstance(port_specific_spec, dict):
            # FIX: 10/3/17 - CHECK HERE THAT, IF MECHANISM ENTRY IS USED, A VARIABLE, WEIGHT AND/OR EXPONENT ENTRY
            # FIX:                       IS APPLIED TO ALL THE OutputPorts SPECIFIED IN OUTPUT_PORTS
            # FIX:                       UNLESS THEY THEMSELVES USE A Port specification dict WITH ANY OF THOSE ENTRIES
            # FIX:           USE ObjectiveMechanism EXAMPLES
            # if MECHANISM in port_specific_spec:
            #     if OUTPUT_PORTS in port_specific_spec
            if SIZE in port_specific_spec:
                if (VARIABLE in port_specific_spec or
                        any(key in port_dict and port_dict[key] is not None for key in {VARIABLE, SIZE})):
                    raise InputPortError(f"PROGRAM ERROR: SIZE specification found in port_specific_spec dict "
                                         f"for {self.__name__} specification of {owner.name} when SIZE or VARIABLE "
                                         f"is already present in its port_specific_spec dict or port_dict.")
                port_dict.update({VARIABLE:np.zeros(port_specific_spec[SIZE])})
                del port_specific_spec[SIZE]
                return port_dict, port_specific_spec
            return None, port_specific_spec

        elif isinstance(port_specific_spec, tuple):

            # GET PORT_SPEC AND ASSIGN PROJECTIONS_SPEC **********************************************************

            tuple_spec = port_specific_spec

            # 2-item tuple specification
            if len(tuple_spec) == 2:

                # 1st item is a value, so treat as Port spec (and return to _parse_port_spec to be parsed)
                #   and treat 2nd item as Projection specification
                if is_numeric(tuple_spec[0]):
                    port_spec = tuple_spec[0]
                    reference_value = port_dict[REFERENCE_VALUE]
                    # Assign value so sender_dim is skipped below
                    # (actual assignment is made in _parse_port_spec)
                    if reference_value is None:
                        port_dict[REFERENCE_VALUE]=port_spec
                    elif not iscompatible(port_spec, reference_value):
                        raise PortError(f"Value in first item of 2-item tuple specification {InputPort.__name__} of "
                                        f"{owner.name} ({port_spec}) is not compatible with its {REFERENCE_VALUE} "
                                        f"({reference_value}).")
                    projections_spec = tuple_spec[1]

                # Tuple is Projection specification that is used to specify the Port,
                else:
                    # return None in port_spec to suppress further, recursive parsing of it in _parse_port_spec
                    port_spec = None
                    if tuple_spec[0] != self:
                        # If 1st item is not the current port (self), treat as part of the projection specification
                        projections_spec = tuple_spec
                    else:
                        # Otherwise, just use 2nd item as projection spec
                        port_spec = None
                        projections_spec = tuple_spec[1]

            # 3- or 4-item tuple specification
            elif len(tuple_spec) in {3,4}:
                # Tuple is projection specification that is used to specify the Port,
                #    so return None in port_spec to suppress further, recursive parsing of it in _parse_port_spec
                port_spec = None
                # Reduce to 2-item tuple Projection specification
                projection_item = tuple_spec[3] if len(tuple_spec)==4 else None
                projections_spec = (tuple_spec[0],projection_item)


            # GET PROJECTIONS IF SPECIFIED *************************************************************************

            try:
                projections_spec
            except UnboundLocalError:
                pass
            else:
                try:
                    params_dict[PROJECTIONS] = _parse_connection_specs(self,
                                                                       owner=owner,
                                                                       connections=projections_spec)
                    # Parse the value of all of the Projections to get/validate variable for InputPort
                    variable = []
                    for projection_spec in params_dict[PROJECTIONS]:
                        # FIX: 10/3/17 - PUTTING THIS HERE IS A HACK...
                        # FIX:           MOVE TO _parse_port_spec UNDER PROCESSING OF ProjectionTuple SPEC
                        # FIX:           USING _get_port_for_socket
                        # from psyneulink.core.components.projections.projection import _parse_projection_spec

                        # Try to get matrix for projection
                        try:
                            sender_dim = np.array(projection_spec.port.value).ndim
                        except AttributeError as e:
                            if (isinstance(projection_spec.port, type) or
                                     projection_spec.port.initialization_status == ContextFlags.DEFERRED_INIT):
                                continue
                            else:
                                raise PortError(f"PROGRAM ERROR: indeterminate value for {projection_spec.port.name} "
                                                f"specified to project to {self.__name__} of {owner.name}.")

                        projection = projection_spec.projection
                        if isinstance(projection, dict):
                            # Don't try to get MATRIX from projection without checking,
                            #    since projection is a defaultDict,
                            #    which will add a matrix entry and assign it to None if it is not there
                            if MATRIX in projection:
                                matrix = projection[MATRIX]
                            else:
                                matrix = None
                        elif isinstance(projection, Projection):
                            if projection.initialization_status == ContextFlags.DEFERRED_INIT:
                                continue
                            # possible needs to be projection.defaults.matrix?
                            matrix = projection.matrix
                        else:
                            raise InputPortError(f"Unrecognized Projection specification for {self.name} of "
                                                 f"{owner.name} ({projection_spec}).")

                        # Determine length of value of projection
                        if matrix is None:
                            # If a reference_value has been specified, it presumably represents the item of the
                            #    owner Mechanism's default_variable to which the InputPort corresponds,
                            #    so use that to constrain the InputPort's variable
                            if port_dict[REFERENCE_VALUE] is not None:
                                variable.append(port_dict[REFERENCE_VALUE])
                                continue
                            # If matrix has not been specified, no worries;
                            #    variable_item can be determined by value of sender
                            sender_shape = np.array(projection_spec.port.value).shape
                            variable_item = np.zeros(sender_shape)
                            # If variable_item HASN'T been specified, or it is same shape as any previous ones,
                            #     use sender's value
                            if ((VARIABLE not in port_dict or port_dict[VARIABLE] is None) and
                                    (not variable or variable_item.shape == variable[0].shape)):
                                # port_dict[VARIABLE] = variable
                                variable.append(variable_item)
                            # If variable HAS been assigned, make sure value is the same for this sender
                            elif np.array(port_dict[VARIABLE]).shape != variable_item.shape:
                                # If values for senders differ, assign None so that Port's default is used
                                variable = None
                                # No need to check any more Projections
                                break

                        # Remove dimensionality of sender OutputPort, and assume that is what receiver will receive
                        else:
                            proj_val_shape = matrix.shape[sender_dim :]
                            # port_dict[VARIABLE] = np.zeros(proj_val_shape)
                            variable.append(np.zeros(proj_val_shape))
                    # Sender's value has not been defined or senders have values of different lengths,
                    if not variable:
                        # If reference_value was provided, use that as the InputPort's variable
                        #    (i.e., assume its function won't transform it)
                        if REFERENCE_VALUE in port_dict and port_dict[REFERENCE_VALUE] is not None:
                            port_dict[VARIABLE] = port_dict[REFERENCE_VALUE]
                        # Nothing to use as variable, so raise exception and allow it to be handled "above"
                        else:
                            raise AttributeError(DEFER_VARIABLE_SPEC_TO_MECH_MSG)
                    else:
                        port_dict[VARIABLE] = variable

                except InputPortError:
                    raise InputPortError(f"Tuple specification in {InputPort.__name__} specification dictionary for "
                                         f"{owner.name} ({projections_spec}) is not a recognized specification for "
                                         f"one or more Mechanisms, {OutputPort.__name__}s, or {Projection.__name__}s "
                                         f"that project to it.")

            # GET WEIGHT AND EXPONENT IF SPECIFIED ***************************************************************

            if len(tuple_spec) == 2:
                pass

            # Tuple is (spec, weights, exponents<, afferent_source_spec>),
            #    for specification of weights and exponents,  + connection(s) (afferent projection(s)) to InputPort
            elif len(tuple_spec) in {3, 4}:

                weight = tuple_spec[WEIGHT_INDEX]
                exponent = tuple_spec[EXPONENT_INDEX]

                if weight is not None and not isinstance(weight, numbers.Number):
                    raise InputPortError(f"Specification of the weight ({weight}) in tuple of {InputPort.__name__} "
                                         f"specification dictionary for {owner.name} must be a number.")
                params_dict[WEIGHT] = weight

                if exponent is not None and not isinstance(exponent, numbers.Number):
                    raise InputPortError(f"Specification of the exponent ({exponent}) in tuple of {InputPort.__name__} "
                                         f"specification dictionary for {owner.name} must be a number.")
                params_dict[EXPONENT] = exponent

            else:
                raise PortError(f"Tuple provided as port_spec for {InputPort.__name__} of {owner.name} "
                                f"({tuple_spec}) must have either 2, 3 or 4 items.")

        elif port_specific_spec is not None:
            raise InputPortError(f"PROGRAM ERROR: Expected tuple or dict for {self.__class__.__name__}-specific params "
                                 f"but, got: {port_specific_spec}.")

        return port_spec, params_dict

    def _parse_self_port_type_spec(self, owner, input_port, context=None):
        """Return InputPort specification dictionary with projections that shadow inputs to input_port

        Called by _parse_port_spec if InputPort specified for a Mechanism belongs to a different Mechanism
        """

        if not isinstance(input_port, InputPort):
            raise InputPortError(f"PROGRAM ERROR: InputPort._parse_self_port_type called "
                                 f"with non-InputPort specification ({input_port}).")

        sender_output_ports = [p.sender for p in input_port.path_afferents]
        port_spec = {NAME: SHADOW_INPUT_NAME + input_port.owner.name,
                      VARIABLE: np.zeros_like(input_port.variable),
                      PORT_TYPE: InputPort,
                      PROJECTIONS: sender_output_ports,
                      PARAMS: {SHADOW_INPUTS: input_port},
                      OWNER: owner}
        return port_spec

    @staticmethod
    def _port_spec_allows_override_variable(spec):
        """
        Returns
        -------
            True - if **spec** outlines a spec for creating an InputPort whose variable can be
                overridden by a default_variable or size argument
            False - otherwise

            ex: specifying an InputPort with a Mechanism allows overriding
        """
        from psyneulink.core.components.mechanisms.mechanism import Mechanism

        if isinstance(spec, Mechanism):
            return True
        if isinstance(spec, collections.abc.Iterable):
            # generally 2-4 tuple spec, but allows list spec
            for item in spec:
                if isinstance(item, Mechanism):
                    return True
                # handles tuple spec where first item of tuple is itself a (name, Mechanism) tuple
                elif (
                    isinstance(item, collections.abc.Iterable)
                    and len(item) >= 2
                    and isinstance(item[1], Mechanism)
                ):
                    return True

        return False

    @property
    def pathway_projections(self):
        return self.path_afferents

    @pathway_projections.setter
    def pathway_projections(self, assignment):
        self.path_afferents = assignment

    @property
    def socket_width(self):
        return self.defaults.variable.shape[-1]

    @property
    def socket_template(self):
        return np.zeros(self.socket_width)

    @property
    def label(self):
        return self.get_label()

    def get_label(self, context=None):
        try:
            label_dictionary = self.owner.input_labels_dict
        except AttributeError:
            label_dictionary = {}
        return self._get_value_label(label_dictionary, self.owner.input_ports, context=context)

    @property
    def position_in_mechanism(self):
        if hasattr(self, "owner"):
            if self.owner is not None:
                return self.owner.get_input_port_position(self)
            else:
                return None
        return None

    @staticmethod
    def _get_port_function_value(owner, function, variable):
        """Put InputPort's variable in a list if its function is LinearCombination and variable is >=2d

        InputPort variable must be embedded in a list so that LinearCombination (its default function)
        returns a variable that is >=2d intact (rather than as arrays to be combined);
        this is normally done in port._update() (and in Port._instantiate-function), but that
        can't be called by _parse_port_spec since the InputPort itself may not yet have been instantiated.

        """

        if (
                (
                    (inspect.isclass(function) and issubclass(function, LinearCombination))
                    or isinstance(function, LinearCombination)
                )
                and isinstance(variable, np.matrix)
        ):
            variable = [variable]

        # if function is None, use Port's default function
        function = function or InputPort.defaults.function

        return Port_Base._get_port_function_value(owner=owner, function=function, variable=variable)


def _instantiate_input_ports(owner, input_ports=None, reference_value=None, context=None):
    """Call Port._instantiate_port_list() to instantiate ContentAddressableList of InputPort(s)

    Create ContentAddressableList of InputPort(s) specified in self.input_ports

    If input_ports is not specified:
        - use owner.input_ports as list of InputPort specifications
        - if owner.input_ports is empty, user owner.defaults.variable to create a default InputPort

    When completed:
        - self.input_ports contains a ContentAddressableList of one or more input_ports
        - self.input_port contains the `primary InputPort <InputPort_Primary>`:  first or only one in input_ports
        - self.input_ports contains the same ContentAddressableList (of one or more input_ports)
        - each InputPort corresponds to an item in the variable of the owner's function
        - the value of all of the input_ports is stored in a list in input_value
        - if there is only one InputPort, it is assigned the full value

    Note: Port._instantiate_port_list()
              parses self.defaults.variable (2D np.array, passed in reference_value)
              into individual 1D arrays, one for each InputPort

    (See Port._instantiate_port_list() for additional details)

    Returns list of instantiated InputPorts
    """

    # This allows method to be called by Mechanism.add_input_ports() with set of user-specified input_ports,
    #    while calls from init_methods continue to use owner.input_ports (i.e., InputPort specifications
    #    assigned in the **input_ports** argument of the Mechanism's constructor)
    input_ports = input_ports or owner.input_ports

    # Parse any SHADOW_INPUTS specs into actual InputPorts to be shadowed
    if input_ports is not None:
        input_ports = _parse_shadow_inputs(owner, input_ports)

    port_list = _instantiate_port_list(owner=owner,
                                         port_list=input_ports,
                                         port_types=InputPort,
                                         port_Param_identifier=INPUT_PORT,
                                         reference_value=reference_value if reference_value is not None
                                                                         else owner.defaults.variable,
                                         reference_value_name=VALUE,
                                         context=context)

    # Call from Mechanism.add_ports, so add to rather than assign input_ports (i.e., don't replace)
    if context.source & (ContextFlags.METHOD | ContextFlags.COMMAND_LINE):
        owner.input_ports.extend(port_list)
    else:
        owner.input_ports = port_list

    # Assign value of require_projection_in_composition
    for port in owner.input_ports:
        # Assign True for owner's primary InputPort and the value has not already been set in InputPort constructor
        if port.require_projection_in_composition is None and owner.input_port == port:
            port.parameters.require_projection_in_composition._set(True, context)

    # Check that number of input_ports and their variables are consistent with owner.defaults.variable,
    #    and adjust the latter if not
    variable_item_is_OK = False
    for i, input_port in enumerate(owner.input_ports):
        try:
            variable_item_is_OK = iscompatible(owner.defaults.variable[i], input_port.defaults.value)
            if not variable_item_is_OK:
                break
        except IndexError:
            variable_item_is_OK = False
            break

    if not variable_item_is_OK:
        old_variable = owner.defaults.variable
        owner.defaults.variable = owner._handle_default_variable(default_variable=[port.defaults.value
                                                                                   for port in owner.input_ports])

        if owner.verbosePref:
            warnings.warn(f"Variable for {old_variable} ({append_type_to_name(owner)}) has been adjusted to match "
                          f"number and format of its input_ports: ({owner.defaults.variable}).")

    return port_list

def _parse_shadow_inputs(owner, input_ports):
    """Parses any {SHADOW_INPUTS:[InputPort or Mechaism,...]} items in input_ports into InputPort specif. dict."""

    input_ports_to_shadow_specs=[]
    for spec_idx, spec in enumerate(input_ports):
        # If {SHADOW_INPUTS:[InputPort or Mechaism,...]} is found:
        if isinstance(spec, dict) and SHADOW_INPUTS in spec:
            input_ports_to_shadow_in_spec=[]
            # For each item in list of items to shadow specified in that entry:
            for item in list(spec[SHADOW_INPUTS]):
                from psyneulink.core.components.mechanisms.mechanism import Mechanism
                # If an InputPort was specified, just used that
                if isinstance(item, InputPort):
                    input_ports_to_shadow_in_spec.append(item)
                # If Mechanism was specified, use all of its InputPorts
                elif isinstance(item, Mechanism):
                    input_ports_to_shadow_in_spec.extend(item.input_ports)
                else:
                    raise InputPortError(f"Specification of {repr(SHADOW_INPUTS)} in for {repr(INPUT_PORTS)} arg of "
                                         f"{owner.name} must be a {Mechanism.__name__} or {InputPort.__name__}.")
            input_ports_to_shadow_specs.append((spec_idx, input_ports_to_shadow_in_spec))

    # If any SHADOW_INPUTS specs were found in input_ports, replace them with actual InputPorts to be shadowed
    if input_ports_to_shadow_specs:
        for item in input_ports_to_shadow_specs:
            idx = item[0]
            del input_ports[idx]
            input_ports[idx:idx] = item[1]
        # Update owner's variable based on full set of InputPorts specified
        owner.defaults.variable, _ = owner._handle_arg_input_ports(input_ports)

    return input_ports



import numpy as np
import psyneulink.core.llvm as pnlvm
import psyneulink.core.components.functions.transferfunctions as Functions
import psyneulink.core.globals.keywords as kw
import pytest

from math import e, pi, sqrt

SIZE=10
test_var = np.random.rand(SIZE)
test_matrix = np.random.rand(SIZE, SIZE)
test_matrix_s = np.random.rand(SIZE, SIZE // 4)
test_matrix_l = np.random.rand(SIZE, 3 * SIZE)

RAND1 = np.random.rand()
RAND2 = np.random.rand()
RAND3 = np.random.rand()
RAND4 = np.random.rand()

softmax_helper = RAND1 * test_var
softmax_helper = softmax_helper - np.max(softmax_helper)
softmax_helper = np.exp(softmax_helper) / np.sum(np.exp(softmax_helper))

tanh_helper = (RAND1 * (test_var + RAND2 - RAND3) + RAND4)
tanh_helper = np.tanh(tanh_helper)

gaussian_helper = e**(-(test_var - RAND2)**2 / (2 * RAND1**2)) / sqrt(2 * pi * RAND1)
gaussian_helper = RAND3 * gaussian_helper + RAND4

def gaussian_distort_helper(seed):
    state = np.random.RandomState([seed])
    # compensate for construction
    state.normal(test_var + RAND1, RAND2)
    return RAND4 * state.normal(test_var + RAND1, RAND2) + RAND3


test_data = [
    (Functions.Linear, test_var, {'slope':RAND1, 'intercept':RAND2}, None, test_var * RAND1 + RAND2),
    (Functions.Exponential, test_var, {'scale':RAND1, 'rate':RAND2}, None, RAND1 * np.exp(RAND2 * test_var)),
    (Functions.Logistic, test_var, {'gain':RAND1, 'x_0':RAND2, 'offset':RAND3, 'scale':RAND4}, None, RAND4 / (1 + np.exp(-(RAND1 * (test_var - RAND2)) + RAND3))),
    (Functions.Tanh, test_var, {'gain':RAND1, 'bias':RAND2, 'x_0':RAND3, 'offset':RAND4}, None, tanh_helper),
    (Functions.ReLU, test_var, {'gain':RAND1, 'bias':RAND2, 'leak':RAND3}, None, np.maximum(RAND1 * (test_var - RAND2), RAND3 * RAND1 *(test_var - RAND2))),
    (Functions.Gaussian, test_var, {'standard_deviation':RAND1, 'bias':RAND2, 'scale':RAND3, 'offset':RAND4}, None, gaussian_helper),
    (Functions.GaussianDistort, test_var.tolist(), {'bias': RAND1, 'variance':RAND2, 'offset':RAND3, 'scale':RAND4 }, None, gaussian_distort_helper(0)),
    (Functions.GaussianDistort, test_var.tolist(), {'bias': RAND1, 'variance':RAND2, 'offset':RAND3, 'scale':RAND4, 'seed':0 }, None, gaussian_distort_helper(0)),
    (Functions.SoftMax, test_var, {'gain':RAND1, 'per_item': False}, None, softmax_helper),
    (Functions.SoftMax, test_var, {'gain':RAND1, 'params':{kw.OUTPUT_TYPE:kw.MAX_VAL}, 'per_item': False}, None, np.where(softmax_helper == np.max(softmax_helper), np.max(softmax_helper), 0)),
    (Functions.SoftMax, test_var, {'gain':RAND1, 'params':{kw.OUTPUT_TYPE:kw.MAX_INDICATOR}, 'per_item': False}, None, np.where(softmax_helper == np.max(softmax_helper), 1, 0)),
    (Functions.LinearMatrix, test_var.tolist(), {'matrix':test_matrix.tolist()}, None, np.dot(test_var, test_matrix)),
    (Functions.LinearMatrix, test_var.tolist(), {'matrix':test_matrix_l.tolist()}, None, np.dot(test_var, test_matrix_l)),
    (Functions.LinearMatrix, test_var.tolist(), {'matrix':test_matrix_s.tolist()}, None, np.dot(test_var, test_matrix_s)),
]

relu_derivative_helper = lambda x : RAND1 if x > 0 else RAND1 * RAND3
logistic_helper = RAND4 / (1 + np.exp(-(RAND1 * (test_var - RAND2)) + RAND3))
tanh_derivative_helper = (RAND1 * (test_var + RAND2) + RAND3)
tanh_derivative_helper = (1 - np.tanh(tanh_derivative_helper)**2) * RAND4 * RAND1
derivative_test_data = [
    (Functions.Linear, test_var, {'slope':RAND1, 'intercept':RAND2}, RAND1),
    (Functions.Exponential, test_var, {'scale':RAND1, 'rate':RAND2}, RAND1 * RAND2 * np.exp(RAND2 * test_var)),
    (Functions.Logistic, test_var, {'gain':RAND1, 'x_0':RAND2, 'offset':RAND3, 'scale':RAND4}, RAND1 * RAND4 * logistic_helper * (1 - logistic_helper)),
    (Functions.ReLU, test_var, {'gain':RAND1, 'bias':RAND2, 'leak':RAND3}, list(map(relu_derivative_helper, test_var))),
    (Functions.Tanh, test_var, {'gain':RAND1, 'bias':RAND2, 'offset':RAND3, 'scale':RAND4}, tanh_derivative_helper),
]

# use list, naming function produces ugly names
names = [
    "LINEAR",
    "EXPONENTIAL",
    "LOGISTIC",
    "TANH",
    "RELU",
    "GAUSIAN",
    "GAUSSIAN DISTORT GLOBAL SEED",
    "GAUSSIAN DISTORT",
    "SOFT_MAX ALL",
    "SOFT_MAX MAX_VAL",
    "SOFT_MAX MAX_INDICATOR",
    "LINEAR_MATRIX SQUARE",
    "LINEAR_MATRIX WIDE",
    "LINEAR_MATRIX TALL",
]

derivative_names = [
    "LINEAR_DERIVATIVE",
    "EXPONENTIAL_DERIVATIVE",
    "LOGISTIC_DERIVATIVE",
    "RELU_DERIVATIVE",
    "TANH_DERIVATIVE",
]

@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, fail, expected", test_data, ids=names)
@pytest.mark.parametrize("mode", [
    'Python',
    pytest.param('LLVM', marks=pytest.mark.llvm),
    pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_execute(func, variable, params, fail, expected, benchmark, mode):
    f = func(default_variable=variable, **params)
    benchmark.group = "TransferFunction " + func.componentName
    if mode == 'Python':
        ex = f
    elif mode == 'LLVM':
        ex = pnlvm.execution.FuncExecution(f).execute
    elif mode == 'PTX':
        ex = pnlvm.execution.FuncExecution(f).cuda_execute
    res = ex(variable)
    assert np.allclose(res, expected)
    if benchmark.enabled:
        benchmark(f.function, variable)


@pytest.mark.function
@pytest.mark.transfer_function
@pytest.mark.benchmark
@pytest.mark.parametrize("func, variable, params, expected", derivative_test_data, ids=derivative_names)
@pytest.mark.parametrize("mode", [
    'Python',
    pytest.param('LLVM', marks=pytest.mark.llvm),
    pytest.param('PTX', marks=[pytest.mark.llvm, pytest.mark.cuda])])
def test_execute_derivative(func, variable, params, expected, benchmark, mode):
    f = func(default_variable=variable, **params)
    benchmark.group = "TransferFunction " + func.componentName + " Derivative"
    if mode == 'Python':
        ex = f.derivative
    elif mode == 'LLVM':
        ex = pnlvm.execution.FuncExecution(f, tags=frozenset({"derivative"})).execute
    elif mode == 'PTX':
        ex = pnlvm.execution.FuncExecution(f, tags=frozenset({"derivative"})).cuda_execute
    res = ex(variable)
    assert np.allclose(res, expected)
    if benchmark.enabled:
        benchmark(ex, variable)


def test_transfer_with_costs_function():
    f = Functions.TransferWithCosts()
    result = f(1)
    assert np.allclose(result, 1)
    f.toggle_cost(Functions.CostFunctions.INTENSITY)
    f = Functions.TransferWithCosts(enabled_cost_functions=Functions.CostFunctions.INTENSITY)
    result = f(2)
    assert np.allclose(result, 2)
    assert np.allclose(f.intensity_cost, 7.38905609893065)
    assert f.adjustment_cost is None
    assert f.duration_cost is None
    assert np.allclose(np.float(f.combined_costs), 7.38905609893065)
    f.toggle_cost(Functions.CostFunctions.ADJUSTMENT)
    result = f(3)
    assert np.allclose(result, 3)
    assert np.allclose(f.intensity_cost, 20.085536923187668)
    assert np.allclose(f.adjustment_cost, 1)
    assert f.duration_cost is None
    assert np.allclose(np.float(f.combined_costs), 21.085536923187668)
    f.toggle_cost(Functions.CostFunctions.DURATION)
    result = f(5)
    assert np.allclose(result, 5)
    assert np.allclose(f.intensity_cost, 148.413159102576603)
    assert np.allclose(f.adjustment_cost, 2)
    assert np.allclose(f.duration_cost, 5)
    assert np.allclose(np.float(f.combined_costs), 155.413159102576603)
    result = f(1)
    assert np.allclose(result, 1)
    assert np.allclose(f.intensity_cost, 2.718281828459045)
    assert np.allclose(f.adjustment_cost, 4)
    assert np.allclose(f.duration_cost, 6)
    assert np.allclose(np.float(f.combined_costs), 12.718281828459045)


@pytest.mark.parametrize(
    'default_variable, func_name, expected_func_variable, expected_func_value',
    [
        ([1, 2, 3], 'transfer_fct', [1, 2, 3], [1, 2, 3])
    ]
)
def test_transfer_with_costs_shapes(
    default_variable,
    func_name,
    expected_func_variable,
    expected_func_value
):
    twc = Functions.TransferWithCosts(default_variable=default_variable)

    np.testing.assert_array_equal(
        getattr(twc.parameters, func_name).get().defaults.variable,
        expected_func_variable
    )
    np.testing.assert_array_equal(
        getattr(twc.parameters, func_name).get().defaults.value,
        expected_func_value
    )



from . import mappingprojection
from . import pathwayprojection

from .mappingprojection import *
from .pathwayprojection import *

__all__ = list(mappingprojection.__all__)
__all__.extend(pathwayprojection.__all__)



import numpy as np
import psyneulink as pnl

# -*- coding: utf-8 -*-
"""Rumelpy_for_Jon

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1Z_v_If7hB9HPzqM4zO51avauyoLwtUIx
"""

# Stimuli and Relations
# [Use for Homework 1a.]

nouns = ['oak', 'pine', 'rose', 'daisy', 'canary', 'robin', 'salmon', 'sunfish']
relations = ['is', 'has', 'can']
is_list = ['living', 'living thing', 'plant', 'animal', 'tree', 'flower', 'bird', 'fish', 'big', 'green', 'red',
           'yellow']
has_list = ['roots', 'leaves', 'bark', 'branches', 'skin', 'feathers', 'wings', 'gills', 'scales']
can_list = ['grow', 'move', 'swim', 'fly', 'breathe', 'breathe underwater', 'breathe air', 'walk', 'photosynthesize']
descriptors = [nouns, is_list, has_list, can_list]

truth_nouns = np.identity(len(nouns))

truth_is = np.zeros((len(nouns), len(is_list)))

truth_is[0, :] = [1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0]
truth_is[1, :] = [1, 1, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0]
truth_is[2, :] = [1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0]
truth_is[3, :] = [1, 1, 1, 0, 0, 1, 0, 0, 0, 0, 0, 0]
truth_is[4, :] = [1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1]
truth_is[5, :] = [1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1]
truth_is[6, :] = [1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 0]
truth_is[7, :] = [1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 0, 0]

truth_has = np.zeros((len(nouns), len(has_list)))

truth_has[0, :] = [1, 1, 1, 1, 0, 0, 0, 0, 0]
truth_has[1, :] = [1, 1, 1, 1, 0, 0, 0, 0, 0]
truth_has[2, :] = [1, 1, 0, 0, 0, 0, 0, 0, 0]
truth_has[3, :] = [1, 1, 0, 0, 0, 0, 0, 0, 0]
truth_has[4, :] = [0, 0, 0, 0, 1, 1, 1, 0, 0]
truth_has[5, :] = [0, 0, 0, 0, 1, 1, 1, 0, 0]
truth_has[6, :] = [0, 0, 0, 0, 0, 0, 0, 1, 1]
truth_has[7, :] = [0, 0, 0, 0, 0, 0, 0, 1, 1]

truth_can = np.zeros((len(nouns), len(can_list)))

truth_can[0, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[1, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[2, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[3, :] = [1, 0, 0, 0, 0, 0, 0, 0, 1]
truth_can[4, :] = [1, 1, 0, 1, 1, 0, 1, 1, 0]
truth_can[5, :] = [1, 1, 0, 1, 1, 0, 1, 1, 0]
truth_can[6, :] = [1, 1, 1, 0, 1, 1, 0, 0, 0]
truth_can[7, :] = [1, 1, 1, 0, 1, 1, 0, 0, 0]

truths = [[truth_nouns], [truth_is], [truth_has], [truth_can]]


def gen_input_vals(nouns, relations):

    rumel_nouns_bias=np.vstack((np.identity(len(nouns)),np.ones((1,len(nouns)))))
    rumel_nouns_bias=rumel_nouns_bias.T

    rumel_rels_bias=np.vstack((np.identity(len(relations)),np.ones((1,len(relations)))))
    rumel_rels_bias=rumel_rels_bias.T
    return (rumel_nouns_bias, rumel_rels_bias)

nouns_onehot, rels_onehot = gen_input_vals(nouns, relations)

r_nouns = np.shape(nouns_onehot)[0]
c_nouns = np.shape(nouns_onehot)[1]
r_rels = np.shape(rels_onehot)[0]
c_rels = np.shape(rels_onehot)[1]

# Build Transfer Mechanisms

#In order to build in biases, we add an extra node to every layer, including the inputs

#For the input layers, we will use linear transfer mechanisms

n_units=16

#This number of hidden units is taken directly from Rumelhart's paper

nouns_in = pnl.TransferMechanism(name="nouns_in",
                                 default_variable=np.zeros(c_nouns)
                                )

rels_in = pnl.TransferMechanism(name="rels_in",
                                default_variable=np.zeros(c_rels)
                               )

#For the hidden layers, we will be using logistic functions

hn = pnl.TransferMechanism(name="hidden_nouns",
                           size=9,
                           function=pnl.Logistic()
                            )

hm = pnl.TransferMechanism(name="hidden_mixed",
                           size=n_units,
                           function=pnl.Logistic()
                            )

out_sig_I = pnl.TransferMechanism(name="sig_outs_I",
                                  size=len(nouns),
                                  function=pnl.Logistic()
                                    )

out_sig_is = pnl.TransferMechanism(name="sig_outs_is",
                                   size=len(is_list),
                                   function=pnl.Logistic()
                                    )

out_sig_has = pnl.TransferMechanism(name="sig_outs_has",
                                    size=len(has_list),
                                    function=pnl.Logistic()
                                    )

out_sig_can = pnl.TransferMechanism(name="sig_outs_can",
                                    size=len(can_list),
                                    function=pnl.Logistic()
                                    )

# Here we create random matrices to connect the mechanisms to each other


map_nouns_hn = pnl.MappingProjection(
                                matrix=np.random.rand(c_nouns,c_nouns),
                                name="map_nouns_hn",
                                sender=nouns_in,
                                receiver=hn
                                )

map_rels_hm = pnl.MappingProjection(
                                matrix=np.random.rand(c_rels,n_units),
                                name="map_rel_hm",
                                sender=rels_in,
                                receiver=hm
                                )

map_hn_hm = pnl.MappingProjection(
                                matrix=np.random.rand(c_nouns,n_units),
                                name="map_hn_hm",
                                sender=hn,
                                receiver=hm
                                )

map_hm_I = pnl.MappingProjection(
                                matrix=np.random.rand(n_units,len(nouns)),
                                name="map_hm_I",
                                sender=hm,
                                receiver=out_sig_I
                                )

map_hm_is = pnl.MappingProjection(
                                matrix=np.random.rand(n_units,len(is_list)),
                                name="map_hm_is",
                                sender=hm,
                                receiver=out_sig_is
                                )

map_hm_has = pnl.MappingProjection(
                                matrix=np.random.rand(n_units,len(has_list)),
                                name="map_hm_has",
                                sender=hm,
                                receiver=out_sig_has
                                )

map_hm_can = pnl.MappingProjection(
                                matrix=np.random.rand(n_units,len(can_list)),
                                name="map_hm_can",
                                sender=hm,
                                receiver=out_sig_can
                                )

#This block of code constructs the network

RumelNet = pnl.AutodiffComposition(
        patience=10,
        min_delta=0.00001,
        learning_rate=1,
        randomize=False
        )
    
    
RumelNet.add_node(nouns_in)
RumelNet.add_node(rels_in)
RumelNet.add_node(hn)
RumelNet.add_node(hm)
RumelNet.add_node(out_sig_I)
RumelNet.add_node(out_sig_is)
RumelNet.add_node(out_sig_has)
RumelNet.add_node(out_sig_can)

RumelNet.add_projection(sender=nouns_in, projection=map_nouns_hn, receiver=hn)
RumelNet.add_projection(sender=rels_in, projection=map_rels_hm, receiver=hm)
RumelNet.add_projection(sender=hn, projection=map_hn_hm, receiver=hm)
RumelNet.add_projection(sender=hm, projection=map_hm_I, receiver=out_sig_I)
RumelNet.add_projection(sender=hm, projection=map_hm_is, receiver=out_sig_is)
RumelNet.add_projection(sender=hm, projection=map_hm_has, receiver=out_sig_has)
RumelNet.add_projection(sender=hm, projection=map_hm_can, receiver=out_sig_can)

RumelNet.show_graph(output_fmt='jupyter')

# We wish to train our network on pairs of inputs: a noun and a relation.
# On each run, we would like to set the targets for the RELEVANT outputs to be
# the associated truth tables and the targets for the IRRELEVANT outputs to be
# a set of "neutral" targets, producing a somewhat "nonsense" response.

# To get a better grasp on this concept:
# Consider the case that a person asks you which out of the "has" list a robin
# posseses. You will respond, likely correctly, with some of those attributes and
# you will not list any of the "can" attributes, because they're not related to
# the question.

# This is a skill we learn early in life. When asked a simple, closed-form
# question in elementary school, we were not rewarded in the same way for giving
# an answer that was related to the question, but did not actually answer it.
# If the question was, "can a canary fly?" and we gave the answer, "a canary is
# a yellow living thing with wings that can breathe air", we wouldn't technically
# be wrong, but we would also not have answered the question.

# We want to train the network in a similar fashion.

# This block of code creates the targets that will be assigned to outputs irrelevant
# to the input pairs.


irrel_is = np.ones((len(nouns), len(is_list))) * .5
irrel_has = np.ones((len(nouns), len(has_list))) * .5
irrel_can = np.ones((len(nouns), len(can_list))) * .5

# This block of code trains the network using a set of three loops. The innermost
# pair of loops takes each noun and creates the appropriate training inputs and outputs associated
# with its "is", "has", and "can" relations. It will also be associated with an
# identity output.

# After constructing the dictionaries, the middle loop, associated with the nouns,
# trains the network on the dictionaries for n_epochs.

# The outermost loop simply repeats the training on each noun for a set number of
# repetitions.

# You are encouraged to experiment with changing the number of repetitions and
# epochs to see how the network learns best.

# You will find that this code takes a few minutes to run. We have placed flags
# in the loops so you can see that it's not stuck.

n_epochs=5
tot_reps=200

for reps in range(tot_reps):
  print('Training rep: ',reps + 1, ' of: ', tot_reps)
  for noun in range(len(nouns)):
    
    inputs_dict = {}

    targets_dict = {}
    targets_dict[out_sig_is] = []
    targets_dict[out_sig_has] = []
    targets_dict[out_sig_can] = []

    inputs_dict[nouns_in] = []
    targets_dict[out_sig_I] = []
    inputs_dict[rels_in] = []

    for i in range(len(relations)):

      if i==0:
        rel = 'is'

        targ_is = truth_is[noun],
        targ_has = irrel_has[noun],
        targ_can = irrel_can[noun],
        
        targ_is=np.reshape(targ_is,np.amax(np.shape(targ_is)))
        targ_has=np.reshape(targ_has,np.amax(np.shape(targ_has)))
        targ_can=np.reshape(targ_can,np.amax(np.shape(targ_can)))

      elif i==1:
        rel = 'has'

        targ_is = irrel_is[noun] ,
        targ_has = truth_has[noun],
        targ_can = irrel_can[noun],
        
        targ_is=np.reshape(targ_is,np.amax(np.shape(targ_is)))
        targ_has=np.reshape(targ_has,np.amax(np.shape(targ_has)))
        targ_can=np.reshape(targ_can,np.amax(np.shape(targ_can)))


      else:
        rel = 'can'

        targ_is = irrel_is[noun] ,
        targ_has = irrel_has[noun],
        targ_can = truth_can[noun],
        
        targ_is=np.reshape(targ_is,np.amax(np.shape(targ_is)))
        targ_has=np.reshape(targ_has,np.amax(np.shape(targ_has)))
        targ_can=np.reshape(targ_can,np.amax(np.shape(targ_can)))


      targets_dict[out_sig_is].append(targ_is)
      targets_dict[out_sig_has].append(targ_has)
      targets_dict[out_sig_can].append(targ_can)

      inputs_dict[nouns_in].append(nouns_onehot[noun])
      targets_dict[out_sig_I].append(truth_nouns[noun])
      inputs_dict[rels_in].append(rels_onehot[i])

    
    result = RumelNet.run(inputs=[{'inputs': inputs_dict,
                                'targets': targets_dict,
                                'epochs': n_epochs}],do_logging=True)

##################### THIS IS WHERE VISUALIZATION CODE STARTS ##################

# As with previous nets, this cell prints the losses of the network over time.
# Make sure you understand why the losses are noisy.

# If you find your loss isn't to your satisfaction, you can run the previous cell again
# to train your network for another batch.

exec_id = RumelNet.default_execution_id
losses = RumelNet.parameters.losses.get(exec_id)

plt.xlabel('Epoch Number')
plt.ylabel('Loss Value')
plt.title('Losses for the Rumelhart Network')
plt.plot(losses)
print('last lost was: ',losses[-1])

# This code block collects all output data from the network and sorts it into logs of
# outputs from the runs on which each output type was relevant.
# This process retains only those outputs for "is" where the network will have been
# training on a noun-is pair, and so on for all noun-relation pairs.

# If you would like to access the outputs for irrelevant outputs (the "has" output
# from an "is" or "can" run, for example) it is also available in the variables
# created in this cell.

data_I=out_sig_I.log.nparray()[1,1]
data_is=out_sig_is.log.nparray()[1,1]
data_has=out_sig_has.log.nparray()[1,1]
data_can=out_sig_can.log.nparray()[1,1]

data_I=np.array(data_I[1][1::])
data_I=np.matrix(data_I)

data_is=np.array(data_is[1][1::])
data_is=np.matrix(data_is)
                 
data_has=np.array(data_has[1][1::])
data_has=np.matrix(data_has)

data_can=np.array(data_can[1][1::])
data_can=np.matrix(data_can)

log_length=np.shape(data_I)[0]

I_rel_log=data_I[0,:].T
is_rel_log=data_is[0,:].T
has_rel_log=data_has[1,:].T
can_rel_log=data_can[2,:].T

for i in range(int(log_length / len(relations)) - 1):
  I_rel_log = np.append(I_rel_log,data_I[3 * (i + 1), :].T, 1)
  is_rel_log = np.append(is_rel_log,data_is[3 * (i + 1), :].T, 1)
  has_rel_log = np.append(has_rel_log,data_has[3 * (i + 1) + 1, :].T, 1)
  can_rel_log = np.append(can_rel_log,data_can[3 * (i + 1) + 2, :].T, 1)
  
  # This cell plots the last output values from the network for each noun/relation pair
# This will show you what the network has learned with regards to the properties of
# each noun.

for i in range(len(nouns)):
  n=-i
  
  plt.stem(I_rel_log[:, n - 1])
  
  plt.title(nouns[n - 1])
  plt.ylabel('Strength of Association')
  plt.xticks(np.arange(len(nouns)), nouns,rotation=35)
  plt.yticks(np.arange(0,1.1,.1))
  plt.show()

  
  plt.stem(is_rel_log[:, n - 1])
  
  plt.title([nouns[n - 1], ' is:'])
  plt.ylabel('Strength of Association')
  plt.xticks(np.arange(len(is_list)), is_list,rotation=35)
  plt.yticks(np.arange(0,1.1,.1))
  plt.show()
  
  plt.stem(has_rel_log[:, n - 1])
  
  plt.title([nouns[n - 1], ' has:'])
  plt.ylabel('Strength of Association')
  plt.xticks(np.arange(len(has_list)), has_list,rotation=35)
  plt.yticks(np.arange(0,1.1,.1))
  plt.show()

  plt.stem(can_rel_log[:, n - 1])
  
  plt.title([nouns[n - 1], ' can:'])
  plt.ylabel('Strength of Association')
  plt.xticks(np.arange(len(can_list)), can_list,rotation=35)
  plt.yticks(np.arange(0,1.1,.1))
  plt.show()
